#-*- coding:utf-8 -*-

#  Copyright © 2009-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
from collections import OrderedDict

from PyQt5.QtCore import Qt, QFileSystemWatcher, QTimer
from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from PyQt5.QtGui import (QPainter, QFontMetrics, QLinearGradient, QPalette, QPen, QBrush, QIcon,
                         QKeySequence, QStandardItemModel, QStandardItem)
from PyQt5.QtWidgets import (QMainWindow, QSizePolicy, QAction, QMessageBox,
                        QLabel, QLineEdit, QPushButton, QTreeView, QFrame, QAbstractItemView)

from .debug import (debug, DEBUG_ROTATE, DEBUG_DRAW, DEBUG_FPS,
                    DEBUG_LIVESHADER, DEBUG_LIVEPLUGINS)
from . import config
from .settings import settings
from . import dialogs
from . import pluginlib
from .model import Model, empty_model
from .gamestate import GameState
from . import drawingarea

try:
    _, ngettext
except NameError:
    debug('gettext not found')
    _ = lambda string: string
    ngettext = lambda singular, plural, cnt: singular if cnt == 1 else plural
    
    
class AnimType: # pylint: disable=R0903
    __slots__ = ()
    NONE, FWD, BWD, KEY, NEW = range(5)
AnimType = AnimType()
    
    
class ElideLabel (QLabel):
    def paintEvent(self, unused_event):
        p = QPainter(self)
        fm = QFontMetrics(self.font())
        rect = self.contentsRect()
        if fm.width(self.text()) > rect.width():
            gradient = QLinearGradient(rect.topLeft(), rect.topRight())
            start = (rect.width() - 2*rect.height()) / rect.width()
            gradient.setColorAt(start, self.palette().color(QPalette.WindowText))
            gradient.setColorAt(1.0, self.palette().color(QPalette.Window))
            pen = QPen()
            pen.setBrush(QBrush(gradient))
            p.setPen(pen)
        p.drawText(self.rect(), Qt.TextSingleLine, self.text())
        
        
class QMoveEdit (QLineEdit):
    nextword = Signal()
    prevword = Signal()
    swapnext = Signal()
    swapprev = Signal()
    
    def keyPressEvent(self, event):
        if event.matches(QKeySequence.MoveToNextWord):
            self.nextword.emit()
        elif event.matches(QKeySequence.MoveToPreviousWord):
            self.prevword.emit()
        elif event.key() == Qt.Key_Right and event.modifiers() == Qt.AltModifier:
            self.swapnext.emit()
        elif event.key() == Qt.Key_Left and event.modifiers() == Qt.AltModifier:
            self.swapprev.emit()
        else:
            QLineEdit.keyPressEvent(self, event)
            
        
class MainWindow (QMainWindow):
    ### init ###
    def __init__(self, opts, **kwargs):
        super().__init__(**kwargs)
        
        if not os.path.exists(config.USER_DATA_DIR):
            os.makedirs(config.USER_DATA_DIR)
            
        try:
            Model.load_index()
        except OSError as e:
            error_message = str(e)
            debug(error_message)
            QTimer.singleShot(1000, lambda: self.error_dialog(error_message))
            
        self.games_file = opts.games_file
        
        # initialize settings
        try:
            version = settings.load(opts.config_file)
        except OSError as e:
            error_message = _('An error occurred while reading the settings:\n'
                            '{error_message}').format(error_message=e)
            debug(error_message)
            settings.load('')
            #FIXME: don't guess, show message if window is visible
            QTimer.singleShot(1000, lambda: self.error_dialog(error_message))
        else:
            if version == 1:
                from . import schema
                schema.migrate_theme_face_N_color_image_mode(settings)
                values = schema.migrate_game_size_blocks_moves_position(settings)
                #XXX: block indices are not the same in previous version
                #if values is not None:
                #    GameState.migrate_1to2(self.games_file, *values)
            
        # Set default format before any widget is created, so that everything uses the same format.
        # To test this use DEBUG_VFPS. The framerate should be then >60.
        drawingarea.CubeArea.set_default_surface_format()
        
        # UI
        self.ui = dialogs.UI(self, 'main.ui')
        from .ui.main import retranslate
        retranslate(self)
        
        self.ui.splitter.setCollapsible(0, False)
        self.ui.splitter.setStretchFactor(0, 1)
        self.ui.splitter.setStretchFactor(1, 0)
        self.ui.label_debug_text.setVisible(DEBUG_DRAW or DEBUG_FPS)
        
        # set icons
        self.setWindowIcon(QIcon(config.APPICON_FILE))
        self.ui.button_edit_exec.setIcon(QIcon.fromTheme('system-run'))
        self.ui.button_edit_clear.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.action_quit.setIcon(QIcon.fromTheme('application-exit'))
        self.ui.action_rewind.setIcon(QIcon.fromTheme('media-seek-backward'))
        self.ui.action_previous.setIcon(QIcon.fromTheme('media-skip-backward'))
        self.ui.action_stop.setIcon(QIcon.fromTheme('media-playback-stop'))
        self.ui.action_play.setIcon(QIcon.fromTheme('media-playback-start'))
        self.ui.action_next.setIcon(QIcon.fromTheme('media-skip-forward'))
        self.ui.action_forward.setIcon(QIcon.fromTheme('media-seek-forward'))
        self.ui.action_add_mark.setIcon(QIcon.fromTheme('list-add'))
        self.ui.action_remove_mark.setIcon(QIcon.fromTheme('list-remove'))
        self.ui.action_preferences.setIcon(QIcon.fromTheme('document-properties'))
        self.ui.action_info.setIcon(QIcon.fromTheme('help-about'))
        self.ui.action_help.setIcon(QIcon.fromTheme('help'))
        def set_shortcut(qobj, key):
            # Workaround, need to keep a reference to keysequence to avoid exception
            keysequence = QKeySequence(key)
            qobj.setShortcut(keysequence)
        set_shortcut(self.ui.action_selectmodel, settings.action.selectmodel)
        set_shortcut(self.ui.action_initial_state, settings.action.initial_state)
        set_shortcut(self.ui.action_reset_rotation, settings.action.reset_rotation)
        set_shortcut(self.ui.action_preferences, settings.action.preferences)
        
        # widgets that are not created with Qt Designer
        self.playbarstate = self.create_toolbar()
        self.edit_moves = QMoveEdit()
        self.edit_moves.setObjectName("edit_moves")
        self.edit_moves.setFrame(False)
        self.ui.layout_moves.insertWidget(1, self.edit_moves)
        self.edit_moves.returnPressed.connect(self.on_edit_moves_returnPressed)
        self.edit_moves.nextword.connect(self.on_edit_moves_nextword)
        self.edit_moves.prevword.connect(self.on_edit_moves_prevword)
        self.edit_moves.swapprev.connect(self.on_edit_moves_swapprev)
        self.edit_moves.swapnext.connect(self.on_edit_moves_swapnext)
        self.cube_area = self.create_cube_area(opts)
        self.cube_area.drop_color.connect(self.on_cubearea_drop_color)
        self.cube_area.drop_file.connect(self.on_cubearea_drop_file)
        self.cube_area.debug_info.connect(self.on_cubearea_debug_info)
        self.setTabOrder(self.edit_moves, self.cube_area)
        self.setTabOrder(self.cube_area, self.ui.box_sidepane)
        self.cube_area.setFocus(Qt.OtherFocusReason)
        self.status_text = ElideLabel()
        p = self.status_text.sizePolicy()
        p.setHorizontalStretch(1)
        p.setHorizontalPolicy(QSizePolicy.Ignored)
        self.status_text.setSizePolicy(p)
        self.ui.statusbar.addWidget(self.status_text)
        # created when needed
        self.progress_dialog = None
        
        # actions that belongs to no widget
        self.action_jump_to_editbar = QAction(self)
        self.action_jump_to_editbar.setObjectName('action_jump_to_editbar')
        self.action_jump_to_editbar.setShortcut(QKeySequence(settings.action.edit_moves))
        self.action_jump_to_editbar.triggered.connect(self.on_action_jump_to_editbar_triggered)
        self.addAction(self.action_jump_to_editbar)
        self.action_edit_cube = QAction(self)
        self.action_edit_cube.setObjectName('action_edit_cube')
        self.action_edit_cube.setShortcut(QKeySequence(settings.action.edit_cube))
        self.action_edit_cube.triggered.connect(self.on_action_edit_cube_triggered)
        self.addAction(self.action_edit_cube)
        
        if DEBUG_LIVESHADER:
            self.filesystemwatcher = QFileSystemWatcher()
            self.filesystemwatcher.addPath(config.SHADER_DIR)
            self.filesystemwatcher.directoryChanged.connect(lambda unused_path: self.cube_area.reload_shader())
        if DEBUG_LIVEPLUGINS:
            from . import plugins
            self.filesystemwatcher_plugins = QFileSystemWatcher()
            self.filesystemwatcher_plugins.addPaths([config.PLUGINS_DIR,
                                                     os.path.dirname(plugins.__file__)])
            def reload_plugins(unused_path):
                import importlib
                for name, module in sys.modules.items():
                    if name.startswith(('pybiklib.plugins.')):
                        importlib.reload(module)
                plugin_group = self.active_plugin_group
                self.add_plugins_to_sidepane()
                plugin_group = min(plugin_group, len(self.plugin_group_widgets)-1)
                if plugin_group < 0:
                    plugin_group = None
                self.set_active_plugin_group(plugin_group)
            self.filesystemwatcher_plugins.directoryChanged.connect(reload_plugins)
        
        # plugins
        self.treestore = QStandardItemModel()
        self.selected_tree_path = None
        self.active_plugin_group = None
        self.plugin_group_widgets = []  # tuples of (button, treeview)
        self.plugin_helper = pluginlib.PluginHelper()
        self.add_plugins_to_sidepane(False)
        
        self.unsolved = False
        self.animtype = AnimType.NONE
        
        # Load window state from settings
        self.ui.action_toolbar.setChecked(settings.window.toolbar)
        self.ui.action_editbar.setChecked(settings.window.editbar)
        self.ui.action_statusbar.setChecked(settings.window.statusbar)
        self.ui.toolbar.setVisible(settings.window.toolbar)
        self.ui.frame_editbar.setVisible(settings.window.editbar)
        self.ui.statusbar.setVisible(settings.window.statusbar)
        self.move_keys = self.get_move_key_dict()
        
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        settings.keystore.error.connect(self.on_settings_error)
        
        # Reload last game
        self.game = GameState()
        self.update_ui()
        
        self.resize(*settings.window.size)
        divider_position = settings.window.divider
        self.show()
        sizes = self.ui.splitter.sizes()
        sizes = [divider_position, sum(sizes)-divider_position]
        self.ui.splitter.setSizes(sizes)
        
        QTimer.singleShot(0, self.postcreate)
        
    def create_cube_area(self, opts):
        cube_area = drawingarea.CubeArea(opts=opts, model=empty_model)
        cube_area.setObjectName('drawingarea')
        sizepolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizepolicy.setHorizontalStretch(0)
        sizepolicy.setVerticalStretch(0)
        cube_area.setSizePolicy(sizepolicy)
        self.ui.verticalLayout.addWidget(cube_area)
        cube_area.animation_ending.connect(self.on_cubearea_animation_ending)
        cube_area.request_rotation.connect(self.on_cubearea_request_rotation)
        cube_area.request_swap_blocks.connect(self.on_cubearea_request_swap_blocks)
        cube_area.request_rotate_block.connect(self.on_cubearea_request_rotate_block)
        glformat = cube_area.format()
        dialogs.PreferencesDialog.sample_buffers = max(glformat.samples(), 1)
        return cube_area
        
    def create_toolbar(self):
        play_actions = [self.ui.action_rewind, self.ui.action_previous,
                        self.ui.action_stop, self.ui.action_play, self.ui.action_next,
                        self.ui.action_forward, self.ui.action_add_mark, self.ui.action_remove_mark]
        for action in play_actions:
            self.ui.toolbar.addAction(action)
        self.ui.toolbar.addSeparator()
        self.ui.toolbar.addAction(self.ui.action_help)
        playbarstate = PlaybarState(play_actions)
        return playbarstate
        
    def postcreate(self):
        self.load_current_game()
        self.add_plugins_to_sidepane()
        
    ### helper functions ###
    
    @staticmethod
    def get_settings():
        # only for pybiktest.utils
        return settings
        
    @staticmethod
    def keyval_from_name(keystr):
        key = QKeySequence('+'.join(k for k in keystr.split('+') if k != 'KP'))
        if key.count() == 0:
            return 0
        key = key[0]
        if 'KP' in keystr.split('+'):
            key |= int(Qt.KeypadModifier)
        return key
        
    def get_move_key_dict(self):
        return {self.keyval_from_name(key): move for (move, key) in settings.draw.accels}
        
    def add_plugins_to_sidepane(self, load=True):
        # remove old plugins
        self.treestore.clear()
        for button, treeview in self.plugin_group_widgets:
            self.ui.layout_sidepane.removeWidget(button)
            button.deleteLater()
            treeview.setModel(None)
            self.ui.layout_sidepane.removeWidget(treeview)
            treeview.deleteLater()
        self.plugin_group_widgets.clear()
        # get a list of all plugins
        if load:
            plugins = self.plugin_helper.load_plugins()
            if self.plugin_helper.error_messages:
                self.error_dialog('\n\n'.join(self.plugin_helper.error_messages))
        else:
            plugins = []
        # convert plugin list to a tree structure
        ptree = OrderedDict()
        for plugin_path, func in plugins:
            subtree = [None, ptree, None]
            for name, transl in plugin_path:
                subtree = subtree[1].setdefault(name, [None, OrderedDict(), transl])
            if plugin_path:
                subtree[0] = func
        # fill treestore with plugins
        def fill_treestore(parent, tree):
            for name, [func, subtree, transl] in tree.items():
                item = QStandardItem(transl)
                item.setData(func)
                parent.appendRow(item)
                fill_treestore(item, subtree)
        fill_treestore(self.treestore, ptree)
        # create widgets in the sidepane to display the plugins
        for i, (name, [func, subtree, transl]) in enumerate(ptree.items()):
            button = QPushButton(self.ui.box_sidepane)
            button.setText(transl)
            button.setFlat(True)
            button.setFocusPolicy(Qt.TabFocus)
            self.ui.layout_sidepane.addWidget(button)
            obsc = lambda i: lambda: self.on_button_sidepane_clicked(i)
            button.clicked.connect(obsc(i))
            
            treeview = QTreeView(self.ui.box_sidepane)
            treeview.setFrameShape(QFrame.NoFrame)
            treeview.setEditTriggers(QAbstractItemView.NoEditTriggers)
            treeview.setUniformRowHeights(True)
            treeview.setAnimated(True)
            treeview.setHeaderHidden(True)
            treeview.hide()
            treeview.setModel(self.treestore)
            index = self.treestore.index(i, 0)
            treeview.setRootIndex(index)
            self.ui.layout_sidepane.addWidget(treeview)
            treeview.activated.connect(self.on_treeview_activated)
            
            self.plugin_group_widgets.append((button, treeview))
        # visibility
        self.set_active_plugin_group(0 if self.plugin_group_widgets else None)
        if load:
            self.hide_incompatible_plugins()
        
    def hide_incompatible_plugins(self):
        def hide_row(treeview, index):
            func_idx = index.data(Qt.UserRole + 1)
            if func_idx is not None:
                try:
                    self.plugin_helper.get_function(self.game.current_state.model, func_idx)
                except pluginlib.PluginModelCompatError:
                    return True
            hide_all = func_idx is None
            rows = self.treestore.rowCount(index)
            for r in range(rows):
                hide = hide_row(treeview, self.treestore.index(r, 0, index))
                treeview.setRowHidden(r, index, hide)
                hide_all = hide_all and hide
            return hide_all
        for i, (button, treeview) in enumerate(self.plugin_group_widgets):
            index = treeview.rootIndex()
            if hide_row(treeview, index):
                button.hide()
                if i == self.active_plugin_group:
                    self.set_active_plugin_group(0)
            else:
                button.show()
        
    def set_active_plugin_group(self, i):
        if self.active_plugin_group is not None:
            unused_button, treeview = self.plugin_group_widgets[self.active_plugin_group]
            treeview.hide()
        if i is not None:
            unused_button, treeview = self.plugin_group_widgets[i]
            treeview.show()
        self.active_plugin_group = i
        
    def load_current_game(self):
        mtype = settings.game.type
        size = settings.games[mtype].size
        model = Model(mtype, size)
        self.game.load(self.games_file, model)
        self.cube_area.set_glmodel_full(model, self.game.current_state.rotations)
        self.update_ui()
        
    def load_game(self, mtype, size):
        self.save_game()
        if self.is_animating():
            return
        assert mtype is not None
        model = Model(mtype, size)
        settings.game.type = model.type
        settings.games[model.type].size = model.size
        self.game.load(self.games_file, model)
        self.unsolved = False
        self.cube_area.set_glmodel_full(model, self.game.current_state.rotations)
        self.update_ui()
        self.hide_incompatible_plugins()
        
    def save_game(self):
        self.game.save(self.games_file)
        
    def new_game(self, mtype=None, size=None, solved=False):
        if self.is_animating():
            return
        self.game.new_game(None, solved)
        self.unsolved = not solved
        self.cube_area.set_glmodel_full(None, self.game.current_state.rotations)
        self.update_ui()
        
    def is_animating(self):
        return self.cube_area.animation_active
        
    def start_animation(self, move_data, stop_after, animtype):
        if move_data:
            blocks = self.game.current_state.identify_rotation_blocks(
                                move_data.axis, move_data.slice)
            self.cube_area.animate_rotation(move_data, blocks, stop_after)
            self.playbarstate.playing(self.game.mark_before)
            self.animtype = animtype
            return True
        return False
        
    def update_statusbar(self):
        '''This function must be called before any action that change the move queue length
        to reflect total count of moves and after each single move'''
        
        if self.cube_area.editing_model:
            mesg = _('Press the Esc key to exit Edit Mode')
            self.unsolved = False
        else:
            current = self.game.move_sequence.current_place
            total = self.game.move_sequence.queue_length
            solved = self.game.current_state.is_solved()
            
            model_text = str(self.game.current_state.model)
            # substitution for {move_text} in statusbar text
            move_text = ngettext("{current} / {total} move",
                                 "{current} / {total} moves",
                                 total).format(current=current, total=total)
            # substitution for {solved_text} in statusbar text
            solved_text = _('solved') if solved else _('not solved')
            
            # statusbar text
            mesg = _('{model_text}, {move_text}, {solved_text}').format(
                            model_text=model_text,
                            move_text=move_text,
                            solved_text=solved_text)
            if self.unsolved and solved:
                self.unsolved = False
                self.show_solved_message()
        self.status_text.setText(mesg)
        
    def update_ui(self):
        # update toolbar
        if self.game.move_sequence.at_start():
            if self.game.move_sequence.at_end():
                self.playbarstate.empty()
            else:
                self.playbarstate.first()
        else:
            if self.game.move_sequence.at_end():
                self.playbarstate.last(self.game.move_sequence.is_mark_after(-1))
            else:
                self.playbarstate.mid(self.game.move_sequence.is_mark_current())
                
        # update formula
        code, pos = self.game.move_sequence.format(self.game.current_state.model)
        self.set_edit_moves(code, pos)
        
        self.update_statusbar()
        
    def show_solved_message(self):
        dialog = QMessageBox(self)
        dialog.setWindowTitle(_(config.APPNAME))
        dialog.setText(_('Congratulations, you have solved the puzzle!'))
        dialog.setIcon(QMessageBox.Information)
        dialog.setStandardButtons(QMessageBox.Close)
        self.cube_area.set_std_cursor()
        QTimer.singleShot(100, dialog.exec_)
        
    def error_dialog(self, message, delay=0):
        dialog = QMessageBox(self)
        dialog.setWindowTitle(_(config.APPNAME))
        dialog.setText(message)
        dialog.setIcon(QMessageBox.Warning)
        dialog.setStandardButtons(QMessageBox.Close)
        self.cube_area.set_std_cursor()
        if delay:
            QTimer.singleShot(delay, dialog.exec_)
        else:
            dialog.exec_()
        
    def set_progress(self, step, message=None, value_max=None):
        if self.progress_dialog is None:
            self.progress_dialog = dialogs.ProgressDialog(parent=self)
        canceled = self.progress_dialog.tick(step, message, value_max)
        from PyQt5.QtWidgets import qApp
        qApp.processEvents()
        return canceled
        
    def end_progress(self):
        if self.progress_dialog is not None:
            self.progress_dialog.done()
            
    ### application handlers ###
    
    def resizeEvent(self, event):
        settings.window.size = event.size().width(), event.size().height()
        sizes = self.ui.splitter.sizes()
        if sum(sizes) > 0:  # ignore the first resize event where sizes==[0,0]
            settings.window.divider = sizes[0]
        QMainWindow.resizeEvent(self, event)
        
    def closeEvent(self, event):
        try:
            dialogs.Dialog.delete()
            self.cube_area.stop()
            settings.disconnect()
            self.save_game()
            settings.flush()
        finally:
            QMainWindow.closeEvent(self, event)
        
    def on_settings_changed(self, key):
        if key == 'draw.accels':
            self.move_keys = self.get_move_key_dict()
            
    def on_settings_error(self, message):
        self.status_text.setText(message)
        
    MODIFIER_MASK = int(Qt.ShiftModifier | Qt.ControlModifier | Qt.KeypadModifier)
    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.cube_area.set_editing_mode(False)
            self.update_ui()
            return
        try:
            move = self.move_keys[event.key() | int(event.modifiers()) & self.MODIFIER_MASK]
        except KeyError:
            event.ignore()
            QMainWindow.keyPressEvent(self, event)
            return
        else:
            if self.is_animating() and self.animtype != AnimType.KEY:
                return
            if not self.is_animating():
                self.game.move_sequence.truncate()
            self.game.move_sequence.parse(move, len(move), self.game.current_state.model)
            self.update_ui()
            if not self.is_animating():
                move_data = self.game.step_next()
                self.start_animation(move_data, stop_after=False, animtype=AnimType.KEY)
            
    ### GL area handlers ###
    
    def on_cubearea_animation_ending(self, last_move):
        self.cube_area.set_transformations(self.game.current_state.rotations)
        
        if not last_move:
            # go again
            self.update_statusbar()
            if self.animtype == AnimType.BWD:
                move_data = self.game.step_back()
            else:
                move_data = self.game.step_next()
            if self.start_animation(move_data, stop_after=False, animtype=self.animtype):
                return
        self.cube_area.animation_end()
        self.update_ui()
            
    def on_cubearea_request_rotation(self, maxis, mslice, mdir):
        self.game.set_next((maxis, mslice, mdir))
        self.update_statusbar()
        move_data = self.game.step_next()
        self.start_animation(move_data, stop_after=False, animtype=AnimType.NEW)
        
    def on_cubearea_request_swap_blocks(self, blockpos, maxis, mslice, mdir):
        self.game.swap_block(blockpos, maxis, mslice, mdir)
        self.cube_area.set_transformations(self.game.current_state.rotations)
        
    def on_cubearea_request_rotate_block(self, blockpos, rdir):
        self.game.rotate_block(blockpos, rdir)
        self.cube_area.set_transformations(self.game.current_state.rotations)
        
    def on_cubearea_drop_color(self, blockpos, facesym, colorname):
        if blockpos < 0:
            settings.theme.bgcolor = colorname
        else:
            colornum = self.game.current_state.get_colornum(blockpos, facesym)
            facekey = self.game.current_state.model.facekeys[colornum]
            settings.theme.faces[facekey].color = colorname
            
    def on_cubearea_drop_file(self, blockpos, facesym, filename):
        colornum = self.game.current_state.get_colornum(blockpos, facesym)
        facekey = self.game.current_state.model.facekeys[colornum]
        settings.theme.faces[facekey].image = filename
        
    def on_cubearea_debug_info(self, mode, info):
        if mode == 'fps':
            text = "FPS %.1f" % info
        elif mode == 'pick':
            pos = self.cube_area.pickdata.blockpos
            cube = self.game.current_state
            idx, rot = cube._blocksr[pos] #FIXME: access to protected member
            color = cube.get_colorsymbol(pos, self.cube_area.pickdata.symbol)
            angle = self.cube_area.pickdata.angle and format(self.cube_area.pickdata.angle, '.1f')
            text = ('idx {idx}, pos {pos} {indices}, rot {rot!r}\n'
                    'face {0.symbol} ({face}), color {color}\n'
                    'axis {0.maxis}, slice {0.mslice}, dir {0.mdir}, angle {angle}\n'
                    'mouse {1[0]}, {1[1]}').format(self.cube_area.pickdata, self.cube_area.mouse_xy,
                    idx=idx, pos=pos, indices=cube.model.cell_indices[pos], rot=rot,
                    color=color, angle=angle, **info)
        else:
            text = 'unknown mode: ' + mode
        self.ui.label_debug_text.setText(text)
        
    ###
    
    @Slot(int, int)
    def on_splitter_splitterMoved(self, pos, index):    # pylint: disable=R0201
        if index == 1:
            settings.window.divider = pos
            
    ### sidepane handlers ###
    
    def on_button_sidepane_clicked(self, i):
        self.set_active_plugin_group(i)
        
    def on_treeview_activated(self, index):
        if self.is_animating():
            return
        func_idx = index.data(Qt.UserRole + 1)
        if func_idx is None:
            return
        if self.game.current_state.model is empty_model:
            return
        try:
            func = self.plugin_helper.get_function(self.game.current_state.model, func_idx)
        except pluginlib.PluginModelCompatError as e:
            self.error_dialog(e.args[0])
            return
        # model accepted, now run the plugin function
        game = self.game.copy()
        try:
            func(game)
        except pluginlib.PluginSolverAbort as e:
            self.error_dialog(e.args[0])
            return
        if game.plugin_mode == 'append':
            # append moves
            self.unsolved = False
            position = self.game.move_sequence.mark_and_extend(game.move_sequence)
            changed = position >= 0
            if changed:
                self.game.goto_next_pos(position)
                changed = 'animate'
        elif game.plugin_mode == 'replace':
            # replace game
            self.game.initial_state = game.initial_state
            self.game.current_state = game.current_state
            self.game.move_sequence = game.move_sequence
            self.unsolved = False
            changed = True
        elif game.plugin_mode == 'challenge':
            # challenge
            self.game.initial_state = game.initial_state
            self.game.current_state = game.initial_state.copy()
            self.game.move_sequence.reset()
            self.unsolved = True
            changed = True
        else:
            assert False
        if changed:
            self.cube_area.set_transformations(self.game.current_state.rotations)
            self.update_ui()
            if changed == 'animate':
                move_data = self.game.step_next()
                self.start_animation(move_data, stop_after=False, animtype=AnimType.NEW)
        self.cube_area.update()
        
    ### action handlers ###
    
    @Slot()
    def on_action_new_triggered(self):
        self.new_game(solved=False)
    @Slot()
    def on_action_new_solved_triggered(self):
        self.new_game(solved=True)
    @Slot()
    def on_action_selectmodel_triggered(self):
        dialog = dialogs.SelectModelDialog.run(parent=self)
        if dialog:
            dialog.response_ok.connect(self.load_game)
    @Slot()
    def on_action_quit_triggered(self):
        self.close()
    @Slot()
    def on_action_preferences_triggered(self):
        dialogs.PreferencesDialog.run(parent=self)
    @Slot(bool)
    def on_action_toolbar_toggled(self, checked):
        self.ui.toolbar.setVisible(checked)
        settings.window.toolbar = checked
    @Slot(bool)
    def on_action_editbar_toggled(self, checked):
        self.ui.frame_editbar.setVisible(checked)
        settings.window.editbar = checked
    @Slot(bool)
    def on_action_statusbar_toggled(self, checked):
        self.ui.statusbar.setVisible(checked)
        settings.window.statusbar = checked
    @Slot()
    def on_action_reset_rotation_triggered(self):
        self.cube_area.reset_rotation()
    @Slot()
    def on_action_help_triggered(self):
        dialogs.HelpDialog.run(parent=self)
    @Slot()
    def on_action_info_triggered(self):
        about = dialogs.AboutDialog(parent=self)
        about.run()
        
    @Slot()
    def on_action_rewind_triggered(self):
        if self.is_animating():
            self.cube_area.stop_requested = True
            if self.animtype == AnimType.BWD:
                self.game.step_next() # undo the already applied move
            self.game.goto_prev_mark()
            self.cube_area.animate_abort()
        else:
            needs_update = self.game.goto_prev_mark()
            self.cube_area.set_transformations(self.game.current_state.rotations)
            if needs_update:
                self.cube_area.update()
                self.update_ui()
    @Slot()
    def on_action_previous_triggered(self):
        if self.is_animating():
            if self.animtype == AnimType.BWD:
                # request another BWD move
                self.cube_area.stop_requested = False
            else:
                self.cube_area.stop_requested = True
                self.game.step_back()
            self.cube_area.animate_abort()
            self.cube_area.stop_requested = True
        else:
            move_data = self.game.step_back()
            self.start_animation(move_data, stop_after=True, animtype=AnimType.BWD)
    @Slot()
    def on_action_stop_triggered(self):
        if self.is_animating():
            if self.cube_area.stop_requested:
                self.cube_area.animate_abort()
            else:
                self.cube_area.stop_requested = True
    @Slot()
    def on_action_play_triggered(self):
        if not self.is_animating():
            move_data = self.game.step_next()
            self.start_animation(move_data, stop_after=False, animtype=AnimType.FWD)
    @Slot()
    def on_action_next_triggered(self):
        if self.is_animating():
            sr = self.cube_area.stop_requested
            if self.animtype == AnimType.BWD:
                self.cube_area.stop_requested = True
                self.game.step_next()
            else:
                self.cube_area.stop_requested = False
            self.cube_area.animate_abort()
            self.cube_area.stop_requested = sr
        else:
            move_data = self.game.step_next()
            self.start_animation(move_data, stop_after=True, animtype=AnimType.FWD)
    @Slot()
    def on_action_forward_triggered(self):
        if self.is_animating():
            self.cube_area.stop_requested = True
            if self.animtype != AnimType.BWD:
                self.game.step_back() # undo the already applied move
            self.game.goto_next_mark()
            self.cube_area.animate_abort()
        else:
            self.game.goto_next_mark()
            self.cube_area.set_transformations(self.game.current_state.rotations)
            self.cube_area.update()
            self.update_ui()
    @Slot()
    def on_action_add_mark_triggered(self):
        self.game.move_sequence.mark_current(True)
        self.update_ui()
    @Slot()
    def on_action_remove_mark_triggered(self):
        self.game.move_sequence.mark_current(False)
        self.update_ui()
        
    @Slot()
    def on_action_initial_state_triggered(self):
        if self.is_animating():
            return
        self.game.set_as_initial_state()
        self.update_ui()
        
    ### edit formula handlers ###
    
    def on_action_jump_to_editbar_triggered(self):
        self.edit_moves.setFocus()
        
    def on_edit_moves_returnPressed(self):
        if self.is_animating():
            self.cube_area.animate_abort(update=False)
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        self.game.set_code(code, pos)
        self.cube_area.set_transformations(self.game.current_state.rotations)
        self.cube_area.update()
        self.update_ui()
        
    def on_edit_moves_nextword(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if cpos <= pos:
            moves.advance()
        unused_code, pos = moves.format(self.game.current_state.model)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_prevword(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if cpos >= pos:
            moves.retard()
        unused_code, pos = moves.format(self.game.current_state.model)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_swapnext(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapnext()
        code, pos = moves.format(self.game.current_state.model)
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    def on_edit_moves_swapprev(self):
        code = self.edit_moves.text()
        pos = self.edit_moves.cursorPosition()
        moves, mpos, cpos = self.game.move_sequence.new_from_code(code, pos,
                                        self.game.current_state.model)
        moves.current_place = mpos
        if pos < cpos:
            moves.retard()
        moves.swapprev()
        code, pos = moves.format(self.game.current_state.model)
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    @Slot()
    def on_button_edit_clear_clicked(self):
        self.edit_moves.setText('')
        self.on_edit_moves_returnPressed()
        
    @Slot()
    def on_button_edit_exec_clicked(self):
        self.on_edit_moves_returnPressed()
        
    def set_edit_moves(self, code, pos):
        self.edit_moves.setText(code)
        self.edit_moves.setCursorPosition(pos)
        
    def on_action_edit_cube_triggered(self):
        self.cube_area.animate_abort(update=False)
        self.game.goto_start()
        if DEBUG_ROTATE:
            self.game.current_state.debug_blocksymbols(allsyms=True)
        self.cube_area.set_transformations(self.game.current_state.rotations)
        self.cube_area.set_editing_mode(True)
        self.update_ui()
        
        
class PlaybarState:
    def __init__(self, play_actions):
        self.play_button_list = play_actions
        self.empty()
        
    def set_toolbar_state(self, sflags, vflags, mark):
        if mark:
            vflags ^= 0b11
        for a in reversed(self.play_button_list):
            a.setEnabled(sflags & 1)
            a.setVisible(vflags & 1)
            sflags >>= 1
            vflags >>= 1
            
    # pylint: disable=C0321,C0326
    def empty(self):          self.set_toolbar_state(0b00000000, 0b11011101, False)
    def first(self):          self.set_toolbar_state(0b00011100, 0b11011101, False)
    def mid(self, mark):      self.set_toolbar_state(0b11011111, 0b11011110, mark)
    def last(self, mark):     self.set_toolbar_state(0b11000011, 0b11011110, mark)
    def playing(self, mark):  self.set_toolbar_state(0b11101100, 0b11101110, mark)
    # pylint: enable=C0321
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import re
import pickle
from array import array

from .debug import DEBUG_DRAW, debug
from .config import MODELS_DIR
from .modelcommon import filebyteorder, epsilon, get_texcoords_range, Vector, epsdigits

try:
    _
except NameError:
    _ = lambda t: t
    

def get_texcoords(vector, normal, minx, maxx, miny, maxy):
    x, y = vector.rotationyx_normal(normal)
    return (x - minx) / (maxx - minx), (y - miny) / (maxy - miny)
    
        
class EmptyModel:
    symbols = ''
    symbolsI = ''
    faces = ''
    default_rotation = (0., 0.)
    
    size = sizes = ()
    blocks = []
    cell_count = 0
    cells_visible_faces = []
    
    def __init__(self, *unused_args):
        pass
        
    def __str__(self):
        return 'no puzzle'
        
    def gl_vertex_data(self, *unused_args):
        return (0, b'', (0, 0, 0, 0, 0, 0), ([], 0, 0), []), [()]
        
        
class Model:
    cache_index = {}
    cache_data = {}
    cache_vectors = []
    
    def __init__(self, mtype, size):
        items = self.cache_index['type'][mtype].items()
        self.type = mtype
        for key, value in items:
            setattr(self, key, value)
        self.mformat = _(self.mformat)
        # size for UI and config, sizes for plugins
        self.size = self.cache_index['normsize'][mtype].get(size, self.defaultsize)
        self.sizes, filename = self.cache_index['sizes'][mtype][self.size]
        
        self.load_data(filename)
        
        self.cell_count = len(self.cell_indices)
        self.rotation_indices = {s:i for i, s in enumerate(self.rotation_symbols)}
        self._inplace_rotations = None
        
    @property
    def inplace_rotations(self):
        if self._inplace_rotations is None:
            # slow and only used in edit mode
            self._inplace_rotations = []
            for cellidx in range(self.cell_count):
                inplace_rotations = []
                for rotsym, blocknum_blocknum2 in self.rotated_position.items():
                    for blocknum, blocknum2 in enumerate(blocknum_blocknum2):
                        if blocknum == blocknum2 == cellidx and rotsym:
                            inplace_rotations.append(rotsym)
                self._inplace_rotations.append(sorted(inplace_rotations, key=len)[:2])
        return self._inplace_rotations
        
    def __str__(self):
        return self.mformat.format(*self.sizes)
        
    @classmethod
    def load_index(cls):
        indexfilename = os.path.join(MODELS_DIR, 'index')
        with open(indexfilename, 'rb') as indexfile:
            cls.cache_index = pickle.load(indexfile)
            
    @staticmethod
    def read_vectors(datafile):
        def isplit3(vals):
            it = iter(vals)
            for x in it:
                try:
                    yield Vector((x, next(it), next(it)))
                except StopIteration:
                    raise ValueError('len must be a multiple of 3')
        vlen = array('I')
        vlen.fromfile(datafile, 1)
        if sys.byteorder != filebyteorder:
            vlen.byteswap()
        vlen = vlen[0]
        vectors = array('f')
        vectors.fromfile(datafile, vlen)
        if sys.byteorder != filebyteorder:
            vectors.byteswap()
        return list(isplit3(vectors))
        
    def load_data(self, filename):
        try:
            data = self.cache_data[self.type][self.sizes]
        except KeyError:
            filename = os.path.join(MODELS_DIR, filename)
            debug('loading model data:', filename)
            with open(filename, 'rb') as datafile:
                data = pickle.load(datafile)
                vectors = self.read_vectors(datafile)
            self.__class__.cache_data = data
            self.__class__.cache_vectors = vectors
            data = data[self.type][self.sizes]
        self.vectors = self.cache_vectors
        for key, value in data.items():
            setattr(self, key, value)
            
    @classmethod
    def from_string(cls, modelstr):
        if modelstr == '*':
            return '*', '*', (), None
        re_model = r'''^(\w+)
                        (?: \s+(\w+)
                            (?: \s*(\W)
                                \s*(\w+)
                                (?: \s*\3
                                    \s*(\w+)
                            )?)?
                            (?:\s+with
                                \s+(.+)
                        )?)?\s*$'''
        match = re.match(re_model, modelstr, re.X)
        if match is None:
            raise ValueError('Invalid model: ' + modelstr)
        #TODO: len(sizes) > 3
        mtype, width, height, depth, exp = match.group(1, 2, 4, 5, 6)
        if mtype not in cls.cache_index['types']:
            raise ValueError('Unknown model type %r' % mtype)
        def convert_if_int(value):
            if value is None:
                return value
            try:
                return int(value)
            except ValueError:
                return value
        sizes = tuple(convert_if_int(s) for s in (width, height, depth))
        return modelstr, mtype, sizes, exp
        
    def norm_symbol(self, sym):
        try:
            return self.normal_rotation_symbols[sym]
        except KeyError:
            new_sym = ''
            for c in sym:
                try:
                    new_sym = self.normal_rotation_symbols[new_sym+c]
                except KeyError:
                    raise ValueError('invalid symbol:', sym)
            return new_sym
            
    def rotations_symbolic_to_index(self, rotations):
        for sym in rotations:
            yield self.rotation_indices[sym]
        
    def blocksym_to_blockpos(self, blocksym):
        for i, faces in enumerate(self.cells_visible_faces):
            #FIXME: different blocks can have same visible faces (eg. edges for 4-Cubes)
            if sorted(faces) == sorted(list(blocksym)):
                return i
        raise ValueError('Invalid block symbol:', blocksym)
        
    def blockpos_to_blocksym(self, blockpos, rotsym):
        #TODO: use cells_visible_faces, see blocksym_to_blockpos
        def axisidx_to_sym(axis, idx):
            if idx <= self.sizes[axis] // 2:
                sym = self.symbols[axis]
            else:
                idx = self.sizes[axis]-1 - idx
                sym = self.symbolsI[axis]
            if idx == 0:
                # skip idx for corners
                return sym, self.face_symbolic_to_face_color(sym, rotsym)
            elif idx == 1 and self.sizes[axis] == 3:
                # for size == 3 there is only one edge
                return '', ''
            else:
                # symbol with index to distinguish edge, but no color because the face is not visible
                return sym + str(idx+1), '?'
        x, y, z = self.cell_indices[blockpos]
        symx, colorsymx = axisidx_to_sym(0, x)
        symy, colorsymy = axisidx_to_sym(1, y)
        symz, colorsymz = axisidx_to_sym(2, z)
        return symx + symy + symz, colorsymx + colorsymy + colorsymz
        
    def face_symbolic_to_face_color(self, face, rot):
        for k, v in self.face_permutations[rot].items():
            if v == face:
                return k
        assert False, (face, rot)
        
    def sym_to_move(self, sym, mslice=-1):
        try:
            maxis = self.symbols.index(sym)
            return maxis, mslice, False
        except ValueError:
            maxis = self.symbolsI.index(sym)
            return maxis, mslice, True
        
    def normalize_complete_moves(self, moves):
        syms = ''.join((self.symbols if not mdir else self.symbolsI)[maxis] for maxis, unused_mslice, mdir in moves)
        syms = self.norm_symbol(syms)
        for sym in syms:
            yield self.sym_to_move(sym)
        
    def rotate_symbolic(self, axis, rdir, block, sym):
        rsym = (self.symbols if not rdir else self.symbolsI)[axis]
        block = self.rotated_position[rsym][block]
        sym = self.norm_symbol(sym + rsym)
        return block, sym
        
    def rotate_move(self, complete_move, move):
        caxis, unused_cslice, cdir = complete_move
        maxis, mslice, mdir = move
        caxissym = (self.symbols if cdir else self.symbolsI)[caxis]
        maxissym = (self.symbols if not mdir else self.symbolsI)[maxis]
        raxissym = self.face_permutations[caxissym][maxissym]
        raxis, unused_rslice, rdir = self.sym_to_move(raxissym)
        if mdir != rdir:
            mslice = self.sizes[raxis] - 1 - mslice
        return raxis, mslice, rdir
        
    def _get_selected_moves(self, cellidx, vector):
        for iaxis, axis in enumerate(self.axes):
            vda = vector.dot(axis)
            if -epsilon < vda < epsilon:
                continue
            mslice = self.cell_indices[cellidx][iaxis]
            if mslice is not None:
                yield iaxis, mslice, vda
            
    def get_selected_moves2(self, cellidx, vert1, vert2):
        res = sorted(self._get_selected_moves(cellidx, vert2 - vert1), key=lambda m: m[2])
        a1, s1, d1 = res[0]
        if len(res) == 1:
            return True, (a1, s1, d1>0)
        a2, s2, d2 = res[-1]
        return False, (a1, s1, d1>0,  a2, s2, d2>0)
        
    def get_selected_move_center(self, cellidx, vector):
        moves = ((ma, ms, md) for ma, ms, md in self._get_selected_moves(cellidx, vector) if ms in [0,self.sizes[ma]-1])
        ma, ms, md = max(moves, key=lambda m: m[2])
        return ma, ms, md>0
        
    @staticmethod
    def triangulate(indices, reverse=False):
        indices = iter(indices)
        first = next(indices)
        prev = next(indices)
        for index in indices:
            yield first
            if reverse:
                yield index
                yield prev
            else:
                yield prev
                yield index
            prev = index
            
    @staticmethod
    def triangulate_barycentric(count, indices, reverse=False):
        if count == 3:
            index1, index2, index3 = indices
            yield index1, (1., 0., 0.)
            index2 = index2, (0., 1., 0.)
            index3 = index3, (0., 0., 1.)
            if reverse:
                yield index3; yield index2
            else:
                yield index2; yield index3
        elif count == 4:
            # +---+  4---3  +-1-+ 
            # | / |  | / |  3 2 1 
            # +---+  1---2  +-3-+ 
            index1, index2, index3, index4 = indices
            index1 = index1, (1., 1., 0.)
            index2 = index2, (0., 1., 0.)
            index3 = index3, (0., 1., 1.)
            index4 = index4, (0., 1., 0.)
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
        elif count == 5:
            #   +---+    4---3    +---+ 
            #  /|  /|   /|  /|   2|  /| 
            # + | / |  5 | / |  + 3 3 2 
            #  \|/  |   \|/  |   1|/  | 
            #   +---+    1---2    +-1-+ 
            index1, index2, index3, index4, index5 = indices
            index1 = index1, (0., 1., 1.)
            index2 = index2, (0., 0., 1.)
            index3 = index3, (1., 0., 1.)
            index4 = index4, (1., 0., 1.)
            index5 = index5, (0., 0., 1.)
            if reverse:
                yield index1; yield index3; yield index2; yield index1; yield index4; yield index3
                yield index1; yield index5; yield index4
            else:
                yield index1; yield index2; yield index3; yield index1; yield index3; yield index4
                yield index1; yield index4; yield index5
        elif count == 6:
            #   +---+      1---6      +---+   
            #  /|\  |\    /|\  |\    1|\  |1  
            # + | \ | +  2 | \ | 5  + 3 \ 3 + 
            #  \|  \|/    \|  \|/    2|  \|2  
            #   +---+      3---4      +---+   
            index1, index2, index3, index4, index5, index6 = indices
            index1a = index1, (0., 1., 1.)
            index1b = index1, (0., 1., 1.)
            index2 = index2, (0., 0., 1.)
            index3 = index3, (1., 0., 1.)
            index4 = index4, (1., 0., 1.)
            index5 = index5, (0., 0., 1.)
            index6 = index6, (0., 1., 1.)
            if reverse:
                yield index1a; yield index3; yield index2; yield index1b; yield index4; yield index3
                yield index1b; yield index6; yield index4; yield index4; yield index6; yield index5
            else:
                yield index1a; yield index2; yield index3; yield index1a; yield index3; yield index4
                yield index1a; yield index4; yield index6; yield index4; yield index5; yield index6
        else:
            assert False, count
            
    def gl_label_vertices(self, polygons, faces, mirror_distance):
        for *cell_vertices, faceno in polygons:
            fd = faces[faceno]
            cell_vertices = [self.vectors[i] for i in cell_vertices]
            normal = self.vectors[self.normals[faceno]]
            color = fd.color
            if fd.imagemode:
                texrange = self.texranges_mosaic[faceno]
            else:
                texrange = get_texcoords_range(cell_vertices, normal)
            texpos = (get_texcoords(v, normal, *texrange) for v in cell_vertices)
            x1,y1,x2,y2 = fd.texrect
            texpos = [(tpx*(x2-x1) + x1, tpy*(y2-y1) + y1) for tpx, tpy in texpos]
            
            for (vertex, tp), baryc in self.triangulate_barycentric(
                                                    len(cell_vertices),
                                                    zip(cell_vertices, texpos)):
                yield vertex, normal, color, tp, baryc
            if mirror_distance is not None:
                offset = normal * mirror_distance
                mirror_normal = -normal
                for (vertex, tp), baryc in self.triangulate_barycentric(
                                                    len(cell_vertices),
                                                    zip(cell_vertices, texpos),
                                                    reverse=True):
                    yield vertex + offset, mirror_normal, color, tp, baryc
                    
    def gl_black_vertices(self, polygons):
        def vngroup(vns):
            vns = iter(vns)
            for v in vns:
                yield v, next(vns)
        for vertices_normals in polygons:
            for ivertex, inormal in self.triangulate(vngroup(vertices_normals)):
                yield self.vectors[ivertex], self.vectors[inormal]
                
    def pickarrowdir(self, maxis, mdir, normal):
        if mdir:
            return Vector(self.axes[maxis]).cross(normal).normalised()
        else:
            return normal.cross(Vector(self.axes[maxis])).normalised()
            
    def _gl_pick_triangles_pickdata_helper(self, cellidx, symbol, normal, vert1, vert2, pickdata):
        def roundeps(val):
            return round(val, epsdigits)
        one, masd = self.get_selected_moves2(cellidx, vert1, vert2)
        if one:
            maxis, mslice, mdir = masd
            arrowdir = self.pickarrowdir(maxis, mdir, normal)
            pickdata.append((maxis, mslice, mdir, cellidx, symbol, arrowdir))
        else:
            maxis1, mslice1, mdir1,  maxis2, mslice2, mdir2 = masd
            sdist1 = roundeps(abs(vert1.dot(self.axes[maxis1]) - self.slice_centers[maxis1][mslice1]))
            sdist2 = roundeps(abs(vert1.dot(self.axes[maxis2]) - self.slice_centers[maxis2][mslice2]))
            if sdist1 <= sdist2:
                arrowdir = self.pickarrowdir(maxis1, mdir1, normal)
                pickdata.append((maxis1, mslice1, mdir1, cellidx, symbol, arrowdir))
            else:
                arrowdir = self.pickarrowdir(maxis2, mdir2, normal)
                pickdata.append((maxis2, mslice2, mdir2, cellidx, symbol, arrowdir))
            sdist1 = roundeps(abs(vert2.dot(self.axes[maxis1]) - self.slice_centers[maxis1][mslice1]))
            sdist2 = roundeps(abs(vert2.dot(self.axes[maxis2]) - self.slice_centers[maxis2][mslice2]))
            if sdist1 < sdist2:
                arrowdir = self.pickarrowdir(maxis1, mdir1, normal)
                pickdata.append((maxis1, mslice1, mdir1, cellidx, symbol, arrowdir))
            else:
                arrowdir = self.pickarrowdir(maxis2, mdir2, normal)
                pickdata.append((maxis2, mslice2, mdir2, cellidx, symbol, arrowdir))
        return one
            
    def _gl_pick_triangulate(self, color1, vertsl, normal, mirror_distance):
        for i, verts in enumerate(vertsl):
            for vert in self.triangulate(verts):
                yield color1 + i, vert
        if mirror_distance is not None:
            mirror_offset = normal * mirror_distance
            vertsl = [[v + mirror_offset for v in verts] for verts in vertsl]
            for i, verts in enumerate(vertsl):
                for vert in self.triangulate(verts, reverse=True):
                    yield color1 + i, vert
                    
    def gl_pick_triangles_center(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, no adjacent visible faces
        maxis, mslice, mdir = self.get_selected_move_center(cellidx, normal)
        color = len(pickdata)
        pickdata.append((maxis, mslice, mdir, cellidx, symbol, None))
        verts = [self.vectors[i] for i in edges_iverts]
        for vert in self.triangulate(verts):
            yield color, vert
        if mirror_distance is not None:
            mirror_offset = normal * mirror_distance
            color = len(pickdata)
            pickdata.append((maxis, mslice, not mdir, cellidx, symbol, None))
            for vert in self.triangulate(verts, reverse=True):
                yield color, vert + mirror_offset
                
    def gl_pick_triangles_edge3(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block of a tetrahedron
        # the first and second vertex are on the edge of the tetrahedron
        vert1, vert2, vert3 = (self.vectors[i] for i in edges_iverts)
        color1 = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3]]
        else:
            v12 = (vert1 + vert2) / 2
            vertsl = [[v12, vert2, vert3], [vert1, v12, vert3]]
        yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
            
    def gl_pick_triangles_edge6(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block and has 6 vertices, rotates 2 slices
        # the first and second vertex are on the edge
        vert1, vert2, vert3, vert4, vert5, vert6 = (self.vectors[i] for i in edges_iverts)
        v12 = (vert1 + vert2) / 2
        v45 = (vert4 + vert5) / 2
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3, vert4, vert5, vert6]]
        else:
            vertsl = [[v12, vert2, vert3, vert4, v45], [vert1, v12, v45, vert5, vert6]]
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                
    def gl_pick_triangles_edge4(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on an edge block, 4 vertices, rotate 1 or 2 slices,
        # the first and second vertex are on the edge
        vert1, vert2, vert3, vert4 = (self.vectors[i] for i in edges_iverts)
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert3, vert4]]
        else:
            #assert ncolors == 2
            ec12 = (vert1 + vert2) / 2
            ec34 = (vert3 + vert4) / 2
            vertsl = [[ec12, vert2, vert3, ec34], [vert1, ec12, ec34, vert4]]
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                
    def gl_pick_triangles_corner34(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on a corner block, 3 or 4 vertices,
        # the second vert is on the corner, the first and third are on the edges
        if len(edges_iverts) == 3:
            vert1, vert2, vert3 = (self.vectors[iv] for iv in edges_iverts)
            vert4 = (vert3 + vert1) / 2
        else:
            vert1, vert2, vert3, vert4 = (self.vectors[iv] for iv in edges_iverts)
        # first triangle
        color = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, vert4]]
        else:
            ec12 = (vert1 + vert2) / 2
            vertsl = [[ec12, vert2, vert4], [vert1, ec12, vert4]]
        # second triangle
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert2, vert3, pickdata)
        if one:
            vertsl.append([vert2, vert3, vert4])
        else:
            ec23 = (vert2 + vert3) / 2
            vertsl.extend([[ec23, vert3, vert4], [vert2, ec23, vert4]])
        yield from self._gl_pick_triangulate(color, vertsl, normal, mirror_distance)
                    
    def gl_pick_triangles_corner5(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # selection_mode: simple, face is on a corner block, 5 vertices,
        # the second vert is on the corner, the first and third are on the edges
        vert1, vert2, vert3, vert4, vert5 = (self.vectors[iv] for iv in edges_iverts)
        vc13 = (vert1 + vert3) / 2
        ec45 = (vert4 + vert5) / 2
        color1 = len(pickdata)
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
        if one:
            vertsl = [[vert1, vert2, ec45, vert5]]
        else:
            vertsl = [[vert1, vert2, vc13], [vert1, vc13, ec45, vert5]]
        one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert2, vert3, pickdata)
        if one:
            vertsl.append([vert2, vert3, vert4, ec45])
        else:
            vertsl.extend([[vc13, vert3, vert4, ec45], [vert2, vert3, vc13]])
        yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
        
    def gl_pick_triangles_any(self, cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata):
        # either selection_mode is quadrant or weird model (size==1 in at least one dimension)
        verts = [self.vectors[i] for i in edges_iverts]
        vc = sum(verts, Vector()) / len(verts)
        vert1 = verts[-1]
        for vert2 in verts:
            color1 = len(pickdata)
            one = self._gl_pick_triangles_pickdata_helper(cellidx, symbol, normal, vert1, vert2, pickdata)
            if one:
                vertsl = [[vc, vert1, vert2]]
            else:
                #assert ncolors == 2
                ec = (vert1 + vert2) / 2
                vertsl = [[vc, vert1, ec], [vc, ec, vert2]]
            yield from self._gl_pick_triangulate(color1, vertsl, normal, mirror_distance)
            vert1 = vert2
                
    def gl_pick_triangles(self, polygons, selection_mode, mirror_distance, pickdata):
        for cellidx, (face, picktype), edges_iverts in polygons:
            symbol = self.faces[face]
            normal = self.vectors[self.normals[face]]
            
            if selection_mode == 0 or picktype == -1:
                yield from self.gl_pick_triangles_any(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
            elif selection_mode == 1:
                if picktype == 0:
                    yield from self.gl_pick_triangles_center(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 1:
                    yield from self.gl_pick_triangles_edge3(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 2:
                    yield from self.gl_pick_triangles_edge4(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 3:
                    yield from self.gl_pick_triangles_corner34(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 5:
                    yield from self.gl_pick_triangles_corner5(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                elif picktype == 4:
                    yield from self.gl_pick_triangles_edge6(cellidx, edges_iverts, symbol, normal, mirror_distance, pickdata)
                else:
                    assert False, (selection_mode, picktype)
            else:
                assert False, (selection_mode, picktype)
                    
    def gl_vertex_data(self, selection_mode, theme, mirror_distance):
        vertices = []
        normals = []
        colors = []
        texpos = []
        baryc = []
        # list of [count vertices, face index]
        cnts_block = []
        faces = [theme.faces[fk] for fk in self.facekeys]
        
        for label_polygons, black_polygons in self.block_polygons:
            cnt = 0
            # label vertices
            for v, n, c, t, b in self.gl_label_vertices(label_polygons, faces, mirror_distance):
                vertices += v
                normals += n
                colors += c
                texpos += t
                baryc += b
                cnt += 1
            # bevel, hidden labels
            for v, n in self.gl_black_vertices(black_polygons):
                vertices += v
                normals += n
                #TODO: colors and texpos unused
                colors += (0,0,0)
                texpos += (0,0)
                baryc += (0,0,0)
                cnt += 1
            cnts_block.append(cnt)
        
        idx_debug = len(vertices) // 3
        if DEBUG_DRAW:
            #FIXME: what if len(self.sizes) < 3 ?
            x = float(-self.sizes[0])
            y = float(-self.sizes[1])
            z = float(-self.sizes[2])
            # axes
            vertices += [x-1, y-1, z,  -x, y-1, z]
            colors += [255, 0, 0,  255, 0, 0]
            vertices += [x-1, y-1, z,  x-1, -y, z]
            colors += [0, 255, 0,  0, 255, 0]
            vertices += [x-1, y-1, z,  x-1, y-1, -z]
            colors += [0, 0, 255,  0, 0, 255]
            # selection (modelview)
            vertices += [0, 0, 0,  1, 1, 1]
            colors += [255, 255, 255,  255, 255, 255]
            # selection (viewport)
            vertices += [0, 0, 0,  1, 1, 0]
            colors += [255, 255, 0,  255, 0, 0]
            
            assert len(cnts_block) == self.cell_count
            assert sum(cnts_block) == idx_debug
            
        pickvertices, pickcolors, pickdata = self.gl_pick_data(selection_mode, mirror_distance)
        cnt_pick = len(pickvertices) // 3
        
        vertices = array('f', vertices).tobytes()
        normals = array('f', normals).tobytes()
        colors = array('B', colors).tobytes()
        texpos = array('f', texpos).tobytes()
        barycentric = array('f', baryc).tobytes()
        pickvertices = array('f', pickvertices).tobytes()
        pickcolors = array('B', pickcolors).tobytes()
        
        vertexdata = vertices + normals + colors + texpos + barycentric + pickvertices + pickcolors
        debug('GL data: %d bytes' % len(vertexdata))
        
        ptrnormals = len(vertices)
        ptrcolors = ptrnormals + len(normals)
        ptrtexpos = ptrcolors + len(colors)
        ptrbarycentric = ptrtexpos + len(texpos)
        ptrpickvertices = ptrbarycentric + len(barycentric)
        ptrpickcolors = ptrpickvertices + len(pickvertices)
        
        return (    self.cell_count,
                    vertexdata,
                    (ptrnormals, ptrcolors, ptrtexpos, ptrbarycentric, ptrpickvertices, ptrpickcolors),
                    (cnts_block, idx_debug, cnt_pick),
                    self.rotation_matrices,
               ), pickdata
                
    def gl_pick_data(self, selection_mode, mirror_distance):
        # Pick TRIANGLES vertices
        vertices = []
        colors = []
        pickdata = [()]  # list of (maxis, mslice, mdir, face, center, cellidx, symbol, arrowdir)
        for col, v in self.gl_pick_triangles(self.pick_polygons, selection_mode, mirror_distance, pickdata):
            vertices += v
            color = [(col>>4) & 0xf0, (col) & 0xf0, (col<<4) & 0xf0]
            colors += color
        assert len(vertices) == len(colors)
        return vertices, colors, pickdata
        
        
empty_model = EmptyModel()
    
    



#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# pylint: disable=C0321

import os
import re
import html
from itertools import zip_longest

from PyQt5.QtCore import QAbstractAnimation, QEvent, QObject, QPropertyAnimation, QSize, QStandardPaths, Qt, QTimer
from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from PyQt5.QtGui import (QColor, QDesktopServices, QFontMetrics, QIcon, QIconEngine, QKeySequence, QPixmap,
                         QStandardItem, QStandardItemModel, QTextDocumentFragment)
from PyQt5.QtWidgets import (QAbstractItemDelegate, QColorDialog, QDialog, QDialogButtonBox, QFileDialog, QLabel,
                         QProgressDialog, QPushButton, QStyledItemDelegate, QTextEdit)

from . import config
from .settings import settings
from .theme import Theme
from .model import Model

try:
    _
except NameError:
    _ = lambda t: t
    

class Dialog (QDialog):
    _instance = None
    _classes = []
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.accepted.connect(self.on_response)
        self._ignore_changed = False
        self.bound = {}
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        
    @classmethod
    def run(cls, **kwargs):
        if cls._instance is None:
            cls._instance = cls(**kwargs)
            ret = cls._instance
            cls._classes.append(cls)
        else:
            ret = None
            cls._instance.move(cls._instance.pos())
        cls._instance.show()
        return ret
        
    @classmethod
    def delete(cls):
        while cls._classes:
            cls_ = cls._classes.pop(0)
            # pylint: disable=W0212
            cls_._instance.close()
            cls_._instance.deleteLater()
            # pylint: enable=W0212
            cls_._instance = None
            
    def bind(self, key, widget, getter, setter, signal):
        if isinstance(getter, str): getter = getattr(widget, getter)
        if isinstance(setter, str): setter = getattr(widget, setter)
        if isinstance(signal, str): signal = getattr(widget, signal)
        setter(getattr(settings, key))
        if signal is not None:
            signal.connect(lambda: self.on_widget_changed(key))
        self.bound[key] = (getter, setter, signal)
        
    @staticmethod
    def bind_reset(settings_key, button):
        button.clicked.connect(lambda: delattr(settings, settings_key))
        
    def on_response(self):
        pass
        
    def on_widget_changed(self, key):
        if self._ignore_changed:
            return
        try:
            getter = self.bound[key][0]
        except KeyError:
            return
        self._ignore_changed = True
        setattr(settings, key, getter())
        self._ignore_changed = False
        
    def on_settings_changed(self, key):
        if self._ignore_changed:
            return
        try:
            setter = self.bound[key][1]
        except KeyError:
            return
        self._ignore_changed = True
        setter(getattr(settings, key))
        self._ignore_changed = False
        
        
class ColorIconEngine (QIconEngine):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.color = 'black'
        
    def paint(self, painter, rect, unused_mode, unused_state):
        painter.fillRect(rect, QColor(self.color))
        
        
class ColorButton (QPushButton):
    color_changed = Signal(str)
    
    def __init__(self, replace_button, **kwargs):
        self._icon = ColorIconEngine()
        parent = replace_button.parentWidget()
        super().__init__(QIcon(self._icon), '', parent, **kwargs)
        
        height = self.iconSize().height()
        self.setIconSize(QSize(height * self.width() // self.height(), height))
        self.setSizePolicy(replace_button.sizePolicy())
        self.setObjectName(replace_button.objectName())
        
        layout = replace_button.parentWidget().layout()
        index = layout.indexOf(replace_button)
        position = layout.getItemPosition(index)
        layout.removeWidget(replace_button)
        layout.addWidget(self, *position)
        replace_button.deleteLater()
        self.clicked.connect(self.on_clicked)
        
    def set_color(self, color):
        self._icon.color = color
        self.update()
        
    def get_color(self):
        return self._icon.color
        
    def on_clicked(self):
        dialog = QColorDialog(self)
        dialog.setCurrentColor(QColor(self.get_color()))
        if dialog.exec_() == QDialog.Accepted:
            color = dialog.currentColor().name()
            self.set_color(color)
            self.color_changed.emit(color)
            
        
class ShortcutEditor (QLabel):
    key_pressed = Signal()
    
    def __init__(self, **kwargs):
        super().__init__(text=_('Press a key …'), **kwargs)
        
        self.setFocusPolicy(Qt.StrongFocus)
        self.setAutoFillBackground(True)
        self.key = None
        
    SAFE_MODIFIER_MASK = Qt.ShiftModifier | Qt.ControlModifier
    IGNORE_KEYS = [Qt.Key_Shift, Qt.Key_Control, Qt.Key_Meta, Qt.Key_Alt, Qt.Key_AltGr,
                   Qt.Key_CapsLock, Qt.Key_NumLock, Qt.Key_ScrollLock]
        
    def keyPressEvent(self, event):
        if event.key() in self.IGNORE_KEYS or event.count() != 1:
            return QLabel.keyPressEvent(self, event)
        mod = event.modifiers() & self.SAFE_MODIFIER_MASK
        key = QKeySequence(event.key() | int(mod)).toString().split('+')
        if  event.modifiers() & Qt.KeypadModifier:
            key.insert(-1, 'KP')
        self.key = '+'.join(key)
        self.key_pressed.emit()
        
        
class ShortcutDelegate (QStyledItemDelegate):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def createEditor(self, parent, unused_option, unused_index):
        editor = ShortcutEditor(parent=parent)
        editor.key_pressed.connect(self.on_editor_key_pressed)
        return editor
        
    def setEditorData(self, editor, index):
        pass
        
    def setModelData(self, editor, model, index):
        if editor.key is not None:
            model.setData(index, editor.key, Qt.DisplayRole)
            
    def on_editor_key_pressed(self):
        editor = self.sender()
        self.commitData.emit(editor)
        self.closeEditor.emit(editor, QAbstractItemDelegate.NoHint)
        
        
class PreferencesDialog (Dialog):
    sample_buffers = 0
        
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.ui = UI(self, 'preferences.ui')
        from .ui.preferences import retranslate
        self.ui = retranslate(self)
        
        self.ui.button_animspeed_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_shader_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_antialiasing_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_mirror_faces_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_movekey_add.setIcon(QIcon.fromTheme('list-add'))
        self.ui.button_movekey_remove.setIcon(QIcon.fromTheme('list-remove'))
        self.ui.button_movekey_reset.setIcon(QIcon.fromTheme('document-revert'))
        self.ui.button_color_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_image_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_background_color_reset.setIcon(QIcon.fromTheme('edit-clear'))
        self.ui.button_mousemode_quad.setIcon(QIcon(os.path.join(config.UI_DIR, 'qt', 'mousemode-quad.png')))
        self.ui.button_mousemode_ext.setIcon(QIcon(os.path.join(config.UI_DIR, 'qt', 'mousemode-ext.png')))
        
        self.ui.buttonBox.button(QDialogButtonBox.Close).setDefault(True)
        self.ui.label_needs_restarted.setVisible(False)
        
        # graphic tab
        self.ui.slider_animspeed.setValue(settings.draw.speed)
        self.ui.slider_animspeed.setRange(*settings.draw.speed_range)
        self.bind('draw.speed', self.ui.slider_animspeed, 'value', 'setValue', 'valueChanged')
        self.bind_reset('draw.speed', self.ui.button_animspeed_reset)
        
        shader_names = {'lighting': _('Lighting'), 'simple': _('Simple')}
        for nick in settings.draw.shader_range:
            self.ui.combobox_shader.addItem(shader_names.get(nick, nick), nick)
        self.ui.combobox_shader.setCurrentIndex(settings.draw.shader)
        self.bind('draw.shader', self.ui.combobox_shader,
                                  'currentIndex', 'setCurrentIndex', 'currentIndexChanged')
        self.bind_reset('draw.shader', self.ui.button_shader_reset)
        
        for text in settings.draw.samples_range:
            self.ui.combobox_samples.addItem(_(text), text)
        self.ui.combobox_samples.setCurrentIndex(settings.draw.samples)
        self.bind('draw.samples', self.ui.combobox_samples,
                                  'currentIndex', 'setCurrentIndex', 'currentIndexChanged')
        self.bind_reset('draw.samples', self.ui.button_antialiasing_reset)
        
        self.ui.spinbox_mirror_faces.setRange(*settings.draw.mirror_distance_range)
        def set_mirror_faces(checked):
            self.ui.checkbox_mirror_faces.setChecked(checked)
            self.ui.spinbox_mirror_faces.setEnabled(checked)
        self.bind('draw.mirror_faces', self.ui.checkbox_mirror_faces, 'isChecked', set_mirror_faces, 'toggled')
        self.bind_reset('draw.mirror_faces', self.ui.button_mirror_faces_reset)
        self.bind('draw.mirror_distance', self.ui.spinbox_mirror_faces, 'value', 'setValue', 'valueChanged')
        self.bind_reset('draw.mirror_distance', self.ui.button_mirror_faces_reset)
        
        html_text = '<html><head/><body><p><span style=" color:#aa0000;">{}</span></p></body></html>'
        html_text = html_text.format(html.escape(_('The program needs to be restarted for the changes to take effect.')))
        self.ui.label_needs_restarted.setText(html_text)
        
        # mouse tab
        def set_selection_mode(unused_mode):
            if settings.draw.selection_nick == 'quadrant':
                self.ui.button_mousemode_quad.setChecked(True)
            elif settings.draw.selection_nick == 'simple':
                self.ui.button_mousemode_ext.setChecked(True)
        self.bind('draw.selection', None, None, set_selection_mode, None)
        
        # keys tab
        self.liststore_movekeys = QStandardItemModel(self)
        self.fill_liststore_movekeys()
        self.ui.listview_movekeys.setModel(self.liststore_movekeys)
        self.liststore_movekeys.itemChanged.connect(self.on_liststore_movekeys_itemChanged)
        self.shortcut_delegate = ShortcutDelegate(parent=self.ui.listview_movekeys)
        self.ui.listview_movekeys.setItemDelegateForColumn(1, self.shortcut_delegate)
        
        # theme tab
        location = QStandardPaths.standardLocations(QStandardPaths.PicturesLocation)
        if not location:
            location = QStandardPaths.standardLocations(QStandardPaths.HomeLocation)
            if not location:
                location = ['']
        self.image_dirname = location[0]
        
        self.button_color = ColorButton(self.ui.button_color)
        self.setTabOrder(self.ui.listview_faces, self.button_color)
        self.button_color.color_changed.connect(self.on_button_color_color_changed)
        self.button_background_color = ColorButton(self.ui.button_background_color)
        self.setTabOrder(self.ui.radiobutton_mosaic, self.button_background_color)
        
        # Only a single color, no picture or pattern
        self.stockfiles = ['']
        self.ui.combobox_image.addItem(_('plain'), '')
        for filename, icon in Theme.textures.get_icons():
            self.stockfiles.append(filename)
            self.ui.combobox_image.addItem(icon, '', filename)
        self.ui.combobox_image.addItem(_('select …'), '/')
        
        self.bind_reset_item('theme.faces.{}.color', self.ui.button_color_reset)
        self.bind_reset_item('theme.faces.{}.image', self.ui.button_image_reset)
        self.bind('theme.bgcolor', self.button_background_color, 'get_color', 'set_color', 'color_changed')
        self.bind_reset('theme.bgcolor', self.ui.button_background_color_reset)
        
        self.facenames = Model.cache_index['facenames']
        self.liststore_faces = QStandardItemModel()
        self.ui.listview_faces.setModel(self.liststore_faces)
        self.ui.listview_faces.selectionModel().currentChanged.connect(self.on_listview_faces_currentChanged)
        self.fill_face_selector()
        
    def fill_face_selector(self):
        for i, (facekey, facename) in enumerate(self.facenames):
            self.liststore_faces.appendRow((QStandardItem(i), QStandardItem(_(facename))))
            filename = settings.theme.faces[facekey].image
            if filename.startswith('/'):
                self.image_dirname = os.path.dirname(filename)
        self.ui.listview_faces.setModelColumn(1)
        #XXX: workaround, listview_faces should automatically set to the correct width
        fm = QFontMetrics(self.ui.listview_faces.font())
        width = max(fm.width(_(fn)) for fk, fn in self.facenames) + 8
        self.ui.listview_faces.setMaximumWidth(width)
        
        active_face = 0
        self.current_facekey = self.facenames[active_face][0]
        index = self.liststore_faces.index(active_face, 1)
        self.ui.listview_faces.setCurrentIndex(index)
        
    def fill_liststore_movekeys(self):
        for move, key in settings.draw.accels:
            self.liststore_movekeys.appendRow([QStandardItem(move), QStandardItem(key)])
        self.liststore_movekeys.setHeaderData(0, Qt.Horizontal, _('Move'))
        self.liststore_movekeys.setHeaderData(1, Qt.Horizontal, _('Key'))
            
    def bind_reset_item(self, settings_key, button):
        def on_clicked():
            delattr(settings, settings_key.format(self.current_facekey))
        button.clicked.connect(on_clicked)
        
    @ staticmethod
    def _accel_mods_to_str(accel_mods):
        accel_str = ''
        for a in accel_mods.value_nicks:
            if accel_str:
                accel_str += '+'
            if a.endswith('-mask'):
                a = a[:-5]
            accel_str += a
        return accel_str
        
    def set_imagefile(self, imagefile):
        index_icon = len(self.stockfiles)
        if imagefile.startswith('/'):
            index = index_icon
            icon = QIcon(imagefile)
        else:
            try:
                index = self.stockfiles.index(imagefile)
            except ValueError:
                index = 0
            icon = QIcon()
        self.ui.combobox_image.setItemIcon(index_icon, icon)
        self.ui.combobox_image.setCurrentIndex(index)
        
    def on_settings_changed(self, key):
        Dialog.on_settings_changed(self, key)
        if self._ignore_changed:
            return
        self._ignore_changed = True
        if key == 'draw.samples':
            visible = (self.sample_buffers != 2**settings.draw.samples > 1)
            self.ui.label_needs_restarted.setVisible(visible)
        elif key == 'draw.accels':
            self.liststore_movekeys.clear()
            self.fill_liststore_movekeys()
        elif key == 'theme.faces.{}.color'.format(self.current_facekey):
            self.button_color.set_color(settings.theme.faces[self.current_facekey].color)
        elif key == 'theme.faces.{}.image'.format(self.current_facekey):
            self.set_imagefile(settings.theme.faces[self.current_facekey].image)
        # Not needed:
        #elif key == 'theme.faces.{}.mode'.format(self.current_facekey):
        self._ignore_changed = False
        
    ###
    
    @Slot(bool)
    def on_checkbox_mirror_faces_toggled(self, checked):
        self.ui.spinbox_mirror_faces.setEnabled(checked)
        
    ### mouse handlers ###
    
    def set_mousemode(self, checked, mode):
        if self._ignore_changed:
            return
        if checked:
            self._ignore_changed = True
            settings.draw.selection_nick = mode
            self._ignore_changed = False
            
    @Slot(bool)
    def on_button_mousemode_quad_toggled(self, checked):
        self.set_mousemode(checked, 'quadrant')
        
    @Slot(bool)
    def on_button_mousemode_ext_toggled(self, checked):
        self.set_mousemode(checked, 'simple')
        
    ### key handlers ###
    
    def get_move_key_list(self):
        move_keys = []
        for i in range(self.liststore_movekeys.rowCount()):
            move, key = [self.liststore_movekeys.item(i, j).data(Qt.DisplayRole) for j in (0, 1)]
            move_keys.append((move, key))
        return move_keys
            
    @Slot()
    def on_button_movekey_add_clicked(self):
        row = self.ui.listview_movekeys.currentIndex().row()
        self._ignore_changed = True
        self.liststore_movekeys.insertRow(row, (QStandardItem(''), QStandardItem('')))
        index = self.liststore_movekeys.index(row, 0)
        self.ui.listview_movekeys.setCurrentIndex(index)
        self.ui.listview_movekeys.edit(index)
        self._ignore_changed = False
        
    @Slot()
    def on_button_movekey_remove_clicked(self):
        row = self.ui.listview_movekeys.currentIndex().row()
        self._ignore_changed = True
        self.liststore_movekeys.takeRow(row)
        settings.draw.accels = self.get_move_key_list()
        self._ignore_changed = False
        
    @Slot()
    def on_button_movekey_reset_clicked(self):  # pylint: disable=R0201
        del settings.draw.accels
        
    def on_liststore_movekeys_itemChanged(self, unused_item):
        self._ignore_changed = True
        settings.draw.accels = self.get_move_key_list()
        self._ignore_changed = False
        
    ### appearance handlers ###
    
    def on_listview_faces_currentChanged(self, current):
        active_face = current.row()
        self.current_facekey = self.facenames[active_face][0]
        self._ignore_changed = True
        self.button_color.set_color(settings.theme.faces[self.current_facekey].color)
        self.set_imagefile(settings.theme.faces[self.current_facekey].image)
        imagemode = settings.theme.faces[self.current_facekey].mode_nick
        if imagemode == 'tiled':
            self.ui.radiobutton_tiled.setChecked(True)
        elif imagemode == 'mosaic':
            self.ui.radiobutton_mosaic.setChecked(True)
        self._ignore_changed = False
        
    def on_button_color_color_changed(self, color):
        if self._ignore_changed:
            return
        self._ignore_changed = True
        settings.theme.faces[self.current_facekey].color = color
        self._ignore_changed = False
        
    @Slot(int)
    def on_combobox_image_activated(self, index):
        if self._ignore_changed:
            return
        try:
            filename = self.stockfiles[index]
        except IndexError:
            filename = QFileDialog.getOpenFileName(self, _("Open Image"), self.image_dirname)
            if isinstance(filename, (tuple, list)):
                filename = filename[0]
            if filename == '':
                # canceled, set the old image
                filename = settings.theme.faces[self.current_facekey].image
                self.set_imagefile(filename)
            else:
                self.image_dirname = os.path.dirname(filename)
        self._ignore_changed = True
        settings.theme.faces[self.current_facekey].image = filename
        self._ignore_changed = False
        
    @Slot(bool)
    def on_radiobutton_tiled_toggled(self, checked):
        self.set_imagemode(checked, 'tiled')
    @Slot(bool)
    def on_radiobutton_mosaic_toggled(self, checked):
        self.set_imagemode(checked, 'mosaic')
        
    def set_imagemode(self, checked, mode):
        if self._ignore_changed:
            return
        if checked:
            self._ignore_changed = True
            settings.theme.faces[self.current_facekey].mode_nick = mode
            self._ignore_changed = False
            
        
class SelectModelDialog (Dialog):
    response_ok = Signal(str, tuple)
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.ui = UI(self, 'model.ui')
        from .ui.model import retranslate
        retranslate(self)
        
        self.mtypes = Model.cache_index['types']
        self.modeldefs = [Model.cache_index['type'][mtype] for mtype in self.mtypes]
        mtype_ranges = lambda mtype: tuple((min(vals), max(vals)) for vals in zip(*Model.cache_index['normsize'][mtype].keys()))
        self.sranges = [mtype_ranges(mtype) for mtype in self.mtypes]
        for i, mtype in enumerate(self.mtypes):
            name = self.modeldefs[i]['name']
            self.ui.combobox_model.addItem(_(name), mtype)
        mtypeidx = self.mtypes.index(settings.game.type)
        self.ui.combobox_model.setCurrentIndex(mtypeidx)
        # Hide items after the dialog size is calculated to avoid later resizing
        QTimer.singleShot(0, lambda: self.on_combobox_model_activated(mtypeidx))
        
    @Slot(int)
    def on_combobox_model_activated(self, index):
        mtype = self.mtypes[index]
        size = settings.games[mtype].size
        try:
            size = Model.cache_index['normsize'][mtype][size]
        except KeyError:
            size = self.modeldefs[index]['defaultsize']
        for label, spin, sizename, srange, svalue in zip_longest(
                        (self.ui.label_width, self.ui.label_heigth, self.ui.label_depth),
                        (self.ui.spin_size1, self.ui.spin_size2, self.ui.spin_size3),
                        self.modeldefs[index]['sizenames'],
                        self.sranges[index],
                        size,
                    ):
            if label is None:
                continue
            label.setText(_(sizename) if sizename is not None else '')
            spin.setVisible(sizename is not None)
            if sizename is not None:
                spin.setRange(*srange)
                spin.setValue(svalue)
                
    def on_response(self):
        index = self.ui.combobox_model.currentIndex()
        size_len = len(self.modeldefs[index]['defaultsize'])
        spins = (self.ui.spin_size1, self.ui.spin_size2, self.ui.spin_size3)[:size_len]
        self.response_ok.emit(
                self.mtypes[index],
                tuple(s.value() for s in spins),
                )
        
        
class ProgressDialog (QProgressDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.canceled_ = False
        self.value_max = 10
        self.value = 0
        self.setWindowModality(Qt.WindowModal)
        self.setMaximum(self.value_max)
        self.setMinimumDuration(0)
        self.setAutoReset(False)
        self.canceled.connect(self.on_canceled)
        
    def on_canceled(self):
        self.canceled_ = True
        self.setLabelText(_('Canceling operation, please wait'))
        
    def tick(self, step, message=None, value_max=None):
        if not self.isVisible():
            self.show()
        if message is not None:
            self.setLabelText(message)
        if value_max is not None:
            self.value_max = value_max
            self.value = 0
            self.setMaximum(value_max)
        if step < 0 or self.value > self.value_max:
            self.setValue(0)
            self.setMaximum(0)
        else:
            self.setValue(self.value)
            self.value += step
        return self.canceled_
        
    def done(self):
        self.canceled_ = False
        self.reset()
        
        
class HelpDialog (Dialog):
    
    helpstrings = [
            # The next strings form the text in the help dialog
            _('Using the mouse to rotate the cube'),
            _('Position the mouse cursor over the puzzle and you will see an arrow '
              'that gives you a hint in which direction the slice under the mouse cursor will be rotated.'),
            _('The left mouse button rotates a single slice of the cube in the direction of the arrow.'),
            _('The right mouse button rotates a single slice of the cube against the direction of the arrow.'),
            _('To rotate the whole cube instead of a single slice press the Ctrl key together with the mouse button.'),
            
            _('Using the keyboard to rotate the cube'),
            _('Make sure the keyboard focus is on the cube area (e.g. click on the background of the cube). '
              'The keys can be configured in the preferences dialog, the default is:'),
            _('Moves the left, right, upper, down, front or back slice clockwise.'),
            _('Moves a slice couterclockwise.'),
            _('Moves the whole cube.'),
            
            _('Other keys and buttons'),
            _('Mouse wheel – Zoom in/out'),
            _('Arrow keys, Left mouse button on the background – Changes the direction of looking at the cube.'),
            _('Moves keyboard focus to the sequence editor above the cube area '
              'where you can edit the move sequence in the notation described below. '
              'Hit enter when done.'),
            
            _('Notation for moves'),
            _('All moves, however they are made, are displayed progressively above the cube area:'),
            _('Moves the left, right, upper, down, front or back slice clockwise.'),
            _('Moves a slice couterclockwise.'),
            _('Moves the whole cube.'),
            _('Moves the first, second or third slice from left clockwise. '
              'The allowed numbers are in the range from 1 to the count of parallel slices. '
              '"l1" is always the same as "l" '
              'and for the classic 3×3×3-Cube "l2" is the same as "r2-" and "l3" is the same as "r-".'),
            _('You can use a space to separate groups of moves.'),
        ]
    helpformat = '''
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html>
<head>
<meta name="qrichtext" content="1" />
<style type="text/css">p, li {{ white-space: pre-wrap; }}</style>
</head>

<body>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>{}</li>
<li>{}</li>
<li>{}</li>
</ul>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>4,6,8,2,5,0 – {}</li>
<li>Shift+4, … – {}</li>
<li>Ctrl+4, … – {}</li>
</ul>

<h3>{}</h3>
<ul>
<li>{}</li>
<li>{}</li>
<li>Ctrl+L – {}</li>
</ul>

<h3>{}</h3>
<p>{}</p>
<ul>
<li>l, r, u, d, f, b – {}</li>
<li>l-, r-, u-, d-, f-, b- – {}</li>
<li>L, R, U, D, F, B – {}</li>
<li>l1, l2, l3 – {}</li>
<li>{}</li>
</ul>

</body></html>
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.ui = UI(self, 'help.ui')
        from .ui.help import retranslate
        self.ui = retranslate(self)
        
        helpstrings = [html.escape(s) for s in self.helpstrings]
        helptext = self.helpformat.format(*helpstrings)
        self.ui.text_help.setHtml(helptext)
        
        
def linkedtext_to_html(text):
    html = QTextDocumentFragment.fromPlainText(text).toHtml()
    html = re.sub(r'&lt;((?:http:|https:|text:).*?)\|&gt;', r'<a href="\1">', html)
    html = re.sub(r'&lt;\|&gt;', r'</a>', html)
    return re.sub(r'&lt;((?:http:|https:).*?)&gt;', r'&lt;<a href="\1">\1</a>&gt;', html)
    
class UI:
    def __init__(self, toplevel, filename):
        self._toplevel = toplevel
        from PyQt5 import uic
        uic.loadUi(os.path.join(config.UI_DIR, 'qt', filename), toplevel)
    def __getattr__(self, attrname):
        if attrname[0] == '_':
            raise AttributeError('{!r} object has no attribute {!r}'.format('UI', attrname))
        else:
            return self._toplevel.findChild(QObject, attrname)
            
        
class AboutDialog (QDialog):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.about = UI(self, 'about.ui')
        from .ui.about import retranslate
        retranslate(self)
        
        self.fill_header()
        self.fill_about_tab()
        self.fill_feedback_tab()
        self.fill_translation_tab()
        self.fill_license_tab()
        
        self.index_tab_about = self.about.tab_widget.indexOf(self.about.tab_about)
        self.index_tab_license = self.about.tab_widget.indexOf(self.about.tab_license)
        
        # About tab animation
        self.scrollbar = self.about.text_translators.verticalScrollBar()
        self.animation = QPropertyAnimation(self.scrollbar, 'value')
        self.animation.setLoopCount(-1)
        
    def fill_header(self):
        pixmap = QPixmap(config.APPICON_FILE)
        self.about.label_icon.setPixmap(pixmap)
        self.about.label_appname.setText(_(config.APPNAME))
        self.about.label_version.setText(config.VERSION)
        self.about.label_description.setText(_(config.SHORT_DESCRIPTION))
        
    def fill_about_tab(self):
        self.about.label_copyright.setText(config.COPYRIGHT)
        html_text = '<html><head/><body><p><a href="{}">{}</a></p></body></html>'.format(
                        config.WEBSITE, _('Pybik project website'))
        self.about.label_website.setText(html_text)
        html_template = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li {{ white-space: pre-wrap; }}
</style></head>
<body>{}</body></html>'''
        html_p_template = '<p align="center">{}</p>'
        html_language_template = '<span style=" font-weight:600;">{}</span>'
        html_person_template = '<a href="{}">{}</a>'
        html_p_items = []
        from .translators import translators
        try:
            import icu
        except ImportError:
            print('PyICU is not installed')
            langname_from_code = lambda code, name: name
        else:
            def langname_from_code(code, name):
                iculocale = icu.Locale.createFromName(code)
                lang = iculocale.getLanguage()
                if icu.Locale.createFromName(lang).getDisplayName() == lang:
                    return name
                return str(iculocale.getDisplayName()) or name
        import locale
        sortfunc = lambda v: locale.strxfrm(v[0])
        def gentranslators():
            for language, langcode, persons in translators:
                language = langname_from_code(langcode, language)
                yield language, persons
        for language, persons in sorted(gentranslators(), key=sortfunc):
            language = html.escape(language)
            html_items = [html_language_template.format(language)]
            for name, link in sorted(persons, key=sortfunc):
                name = html.escape(name)
                html_items.append(html_person_template.format(link, name))
            html_p_items.append(html_p_template.format('<br/>'.join(html_items)))
        html_text = html_template.format(''.join(html_p_items))
        self.about.text_translators.setHtml(html_text)
        qdesktopservices_openurl = QDesktopServices.openUrl
        self.about.text_translators.anchorClicked.connect(qdesktopservices_openurl)
        self.about.text_translators.viewport().installEventFilter(self)
        
    def fill_feedback_tab(self):
        html = linkedtext_to_html(config.get_filebug_text())
        self.about.label_feedback.setText(html)
        
    def fill_translation_tab(self):
        html = linkedtext_to_html(_(config.TRANSLATION_TEXT))
        self.about.label_translation.setText(html)
        
    def fill_license_tab(self):
        self.about.text_license_short.hide()
        self.about.text_license_full.hide()
        html = linkedtext_to_html('\n\n'.join((_(config.LICENSE_INFO), _(config.LICENSE_FURTHER))))
        self.about.text_license_short.setHtml(html)
        try:
            with open(config.LICENSE_FILE, 'rt', encoding='utf-8') as license_file:
                text = license_file.read()
        except Exception as e:
            print('Unable to find license text:', e)
            text = _(config.LICENSE_NOT_FOUND)
            self.about.text_license_full.setLineWrapMode(QTextEdit.WidgetWidth)
        html = linkedtext_to_html(text)
        self.about.text_license_full.setHtml(html)
        self.about.tab_widget.currentChanged.connect(self._on_tab_widget_currentChanged)
        self.about.text_license_short.anchorClicked.connect(self._on_text_license_anchorClicked)
        
    def update_animation(self):
        smin = self.scrollbar.minimum()
        smax = self.scrollbar.maximum()
        if smax <= smin:
            return
        self.animation.setDuration((smax-smin) * 40)
        self.animation.setKeyValueAt(0., smin)
        self.animation.setKeyValueAt(0.04, smin)
        self.animation.setKeyValueAt(0.50, smax)
        self.animation.setKeyValueAt(0.54, smax)
        self.animation.setKeyValueAt(1., smin)
        
    def showEvent(self, unused_event):
        if self.animation.state() == QAbstractAnimation.Stopped:
            self.update_animation()
            self.scrollbar.hide()
            self.animation.start()
            
    def resizeEvent(self, unused_event):
        self.update_animation()
        
    def eventFilter(self, unused_obj, event):
        #assert obj == self.about.text_translators.viewport()
        if event.type() in [QEvent.MouseButtonPress, QEvent.Wheel]:
            self.animation.pause()
            self.scrollbar.show()
        return False
        
    def _on_tab_widget_currentChanged(self, index):
        if index == self.index_tab_about:
            self.animation.resume()
            self.scrollbar.hide()
        else:
            self.animation.pause()
            self.scrollbar.show()
        if index == self.index_tab_license:
            self.about.text_license_short.setVisible(True)
            self.about.text_license_full.setVisible(False)
        
    def _on_text_license_anchorClicked(self, url):
        if url.toString() == 'text:FULL_LICENSE_TEXT':
            self.about.text_license_short.setVisible(False)
            self.about.text_license_full.setVisible(True)
        else:
            QDesktopServices.openUrl(url)
            
    def run(self):
        self.exec_()
        self.deleteLater()
        




#-*- coding:utf-8 -*-
# cython: profile=False

#  Copyright © 2009-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False

__all__ = ['ATTRIB_LOCATION', 'PICKATTRIB_LOCATION',
           'matrix_set_identity', 'mat4',
           'gl_draw_cube', 'gl_pick_cube', 'gl_init_buffers', 'gl_delete_buffers',
           'gl_draw_cube_debug', 'gl_draw_select_debug', 'gl_init_object_location']

#px/from libc.math cimport M_PI, cos, sin
from math import radians, cos, sin
#pxd/from gl cimport *
from OpenGL.GL import *     # pylint: disable=W0614,W0401
#px:
import ctypes
from ctypes import sizeof
from OpenGL.raw.GL.VERSION.GL_2_0 import glVertexAttribPointer
#px.

from .debug import debug, DEBUG


if DEBUG:
    debug('Importing module:', __name__)
    debug('  from package:', __package__)
    debug('  compiled:', __compiled)


#pxd/cdef enum: #
if True:
    MAX_TRANSFORMATIONS = 24
    MAX_BLOCKS = 1000
    #pxd:
    ATTRIB_LOCATION = 0
    PICKATTRIB_LOCATION = 5
    #px.

#pxd>ctypedef float vec4[4]
#pxd>ctypedef vec4 mat4[4]
#px-
mat4 = lambda: [[0.]*4 for _i in range(4)]

#pxd/cdef matrix_set_identity(mat4& matrix):
def matrix_set_identity(matrix):
    matrix[0][0] = 1.; matrix[0][1] = 0.; matrix[0][2] = 0.; matrix[0][3] = 0.
    matrix[1][0] = 0.; matrix[1][1] = 1.; matrix[1][2] = 0.; matrix[1][3] = 0.
    matrix[2][0] = 0.; matrix[2][1] = 0.; matrix[2][2] = 1.; matrix[2][3] = 0.
    matrix[3][0] = 0.; matrix[3][1] = 0.; matrix[3][2] = 0.; matrix[3][3] = 1.
    
#px/cdef struct Block:
class Block:     # pylint: disable=R0903
#px-
    def __init__(self):
        #px/vec4 *transformation
        self.transformation = None
        
        #px/bint in_motion
        self.in_motion = None
        
        #px/int idx_triangles
        self.idx_triangles = None
        #px/int cnt_triangles
        self.cnt_triangles = None
        
#px/cdef struct Cube:
class cube:     # pylint: disable=W0232, R0903
    #px/mat4 transformations[MAX_TRANSFORMATIONS]
    transformations = [[[None]*4, [None]*4, [None]*4, [None]*4] for __t in range(MAX_TRANSFORMATIONS)]
        
    #px+unsigned int number_blocks
    #px/Block blocks[MAX_BLOCKS]
    blocks = [Block() for __block in range(MAX_BLOCKS)]
    
    #px+int cnt_pick
    #px+int idx_debug
    
    #px+GLuint object_location
    #px+GLuint glbuffer
    
#px+cdef Cube cube

#px/cdef struct Animation_Struct:
class animation:
    #px+float angle, angle_max
    #px+float rotation_x, rotation_y, rotation_z
    #px/mat4 rotation_matrix
    rotation_matrix = mat4()
#px+cdef Animation_Struct animation


def init_module():
    #px+cdef int i
    
    cube.number_blocks = 0
    #cube.blocks
    
    cube.cnt_pick = 0
    cube.idx_debug = 0
    
    animation.angle = animation.angle_max = 0

def set_transformations(blocks):
    #px+cdef unsigned int i, b
    for i, b in enumerate(blocks):
        cube.blocks[i].transformation = cube.transformations[b]
        cube.blocks[i].in_motion = False
    
#px/cpdef gl_set_atlas_texture(int x, int w, int h, bytes data):
def gl_set_atlas_texture(x, w, h, data):
    #px/glTexSubImage2D(GL_TEXTURE_2D, 0, x, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, <char*>data)
    glTexSubImage2D(GL_TEXTURE_2D, 0, x, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, data)
    
#px/cpdef set_animation_start(blocks, float angle_max, float axisx, float axisy, float axisz):
def set_animation_start(blocks, angle_max, axisx, axisy, axisz):
    animation.angle = 0.0
    animation.angle_max = angle_max
    animation.rotation_x, animation.rotation_y, animation.rotation_z = axisx, axisy, axisz
    matrix_set_identity(animation.rotation_matrix)
    #px+cdef int block_id
    #px+cdef bint selected
    for block_id, unused_block in blocks:
        cube.blocks[block_id].in_motion = True
    
#px/cdef void _update_animation_rotation_matrix():
def _update_animation_rotation_matrix():
    #px/cdef float angle = animation.angle / 180. * M_PI
    angle = radians(animation.angle)
    #px+cdef float x, y, z
    x = animation.rotation_x
    y = animation.rotation_y
    z = animation.rotation_z
    #px+cdef float c, s
    c = cos(angle)
    s = sin(angle)
    
    animation.rotation_matrix[0][0] = x*x*(1-c) + c
    animation.rotation_matrix[0][1] = x*y*(1-c) + z*s
    animation.rotation_matrix[0][2] = x*z*(1-c) - y*s
    animation.rotation_matrix[1][0] = y*x*(1-c) - z*s
    animation.rotation_matrix[1][1] = y*y*(1-c) + c
    animation.rotation_matrix[1][2] = y*z*(1-c) + x*s
    animation.rotation_matrix[2][0] = x*z*(1-c) + y*s
    animation.rotation_matrix[2][1] = y*z*(1-c) - x*s
    animation.rotation_matrix[2][2] = z*z*(1-c) + c
    
#px/cpdef set_animation_next(float increment):
def set_animation_next(increment):
    animation.angle -= increment
    _update_animation_rotation_matrix()
    return abs(animation.angle) < animation.angle_max

#px/cdef void _mult_matrix(mat4 &dest, mat4 &src1, mat4 &src2):
def _mult_matrix(dest, src1, src2):
    #px+cdef int i, j, k
    #px+cdef float sum_
    for j in range(4):
        for i in range(4):
            sum_ = 0.
            for k in range(4):
                sum_ += src1[k][i] * src2[j][k]
            dest[j][i] = sum_
            
#pxd/cdef void gl_draw_cube():
def gl_draw_cube():
    #px+cdef unsigned int i
    #px/cdef mat4 object_matrix
    object_matrix = mat4()
    for i in range(cube.number_blocks):
        if cube.blocks[i].in_motion:
            _mult_matrix(object_matrix, animation.rotation_matrix, cube.blocks[i].transformation)
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
        else:
            #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &cube.blocks[i].transformation[0][0])
            glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, cube.blocks[i].transformation)
        glDrawArrays(GL_TRIANGLES, cube.blocks[i].idx_triangles, cube.blocks[i].cnt_triangles)
        
#pxd/cdef void gl_pick_cube():
def gl_pick_cube():
    glDrawArrays(GL_TRIANGLES, 0, cube.cnt_pick)
    
#pxd/cdef void gl_init_buffers():
def gl_init_buffers():
    #px/glGenBuffers(1, &cube.glbuffer)
    cube.glbuffer = glGenBuffers(1)
    
#pxd/cdef void gl_delete_buffers():
def gl_delete_buffers():
    glBindBuffer(GL_ARRAY_BUFFER, 0)
    #px/glDeleteBuffers(1, &cube.glbuffer)
    glDeleteBuffers(1, [cube.glbuffer])
    cube.glbuffer = 0
    
#px/cdef void _gl_set_pointer(GLuint index, GLint size, GLenum type, GLboolean normalized, long pointer):
def _gl_set_pointer(index, size, type, normalized, pointer):
    #px/glVertexAttribPointer(index, size, type, normalized, 0, <void*>pointer)
    glVertexAttribPointer(index, size, type, normalized, 0, ctypes.cast(pointer, ctypes.c_void_p))
    glEnableVertexAttribArray(index)
    
#px/cpdef gl_set_data(int nblocks, bytes vertexdata, vertexpointers, vertexinfo, transformations):
def gl_set_data(nblocks, vertexdata, vertexpointers, vertexinfo, transformations):
    assert nblocks <= MAX_BLOCKS, '{} blocks, hardcoded limit is {}'.format(
                                  nblocks, MAX_BLOCKS)
    cube.number_blocks = nblocks
    
    #### Create the raw GL data ####
    #px+cdef long normalpointer, colorpointer, texpospointer, barycpointer
    #px+cdef unsigned int idx_debug, cnt_pick
    normalpointer, colorpointer, texpospointer, barycpointer, pickvertpointer, pickcolorpointer = vertexpointers
    cnts_block, idx_debug, cnt_pick = vertexinfo
    
    #px+cdef unsigned int idx_block, idx, i, j
    idx_block = 0
    
    for idx in range(cube.number_blocks):
        cube.blocks[idx].idx_triangles = idx_block
        cube.blocks[idx].cnt_triangles = cnts_block[idx]
        idx_block += cnts_block[idx]
        
    cube.cnt_pick = cnt_pick
    cube.idx_debug = idx_debug
    
    glBindBuffer(GL_ARRAY_BUFFER, cube.glbuffer)
    #px/glBufferData(GL_ARRAY_BUFFER, len(vertexdata), <char*>vertexdata, GL_STATIC_DRAW)
    glBufferData(GL_ARRAY_BUFFER, len(vertexdata), vertexdata, GL_STATIC_DRAW)
    
    _gl_set_pointer(ATTRIB_LOCATION,   3, GL_FLOAT, GL_FALSE, 0)
    _gl_set_pointer(ATTRIB_LOCATION+1, 3, GL_FLOAT, GL_FALSE, normalpointer)
    _gl_set_pointer(ATTRIB_LOCATION+2, 3, GL_UNSIGNED_BYTE, GL_TRUE, colorpointer)
    _gl_set_pointer(ATTRIB_LOCATION+3, 2, GL_FLOAT, GL_FALSE, texpospointer)
    _gl_set_pointer(ATTRIB_LOCATION+4, 3, GL_FLOAT, GL_FALSE, barycpointer)
    _gl_set_pointer(PICKATTRIB_LOCATION, 3, GL_FLOAT, GL_FALSE, pickvertpointer)
    _gl_set_pointer(PICKATTRIB_LOCATION+1, 3, GL_UNSIGNED_BYTE, GL_TRUE, pickcolorpointer)
    
    assert len(transformations) <= MAX_TRANSFORMATIONS, '{} transformations, hardcoded limit is {}'.format(
                                                        len(transformations), MAX_TRANSFORMATIONS)
    for idx in range(len(transformations)):
        for i in range(4):
            for j in range(4):
                cube.transformations[idx][i][j] = float(transformations[idx][i][j])
                
#pxd/cdef void gl_draw_cube_debug():
def gl_draw_cube_debug():
    #px/cdef mat4 object_matrix
    object_matrix = mat4()
    matrix_set_identity(object_matrix)
    #px/glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, &object_matrix[0][0])
    glUniformMatrix4fv(cube.object_location, 1, GL_FALSE, object_matrix)
    glDrawArrays(GL_LINES, cube.idx_debug, 6)
    
#pxd/cdef void gl_draw_select_debug(GLfloat *selectdata, GLsizeiptr size, GLuint prog_hud):
def gl_draw_select_debug(selectdata, size, prog_hud):
    #px+cdef int i, j
    #px+cdef GLintptr offset
    offset = (cube.idx_debug+6) * 3 * sizeof(GLfloat)
    #px/glBufferSubData(GL_ARRAY_BUFFER, offset, size, &selectdata[0])
    glBufferSubData(GL_ARRAY_BUFFER, offset, len(selectdata) * sizeof(GLfloat), ArrayDatatype.asArray(selectdata, GL_FLOAT))
    
    glDisable(GL_DEPTH_TEST)
    glDrawArrays(GL_LINES, cube.idx_debug+6, 2)
    glUseProgram(prog_hud)
    glDrawArrays(GL_POINTS, cube.idx_debug+6+2, 2)
    glEnable(GL_DEPTH_TEST)
    
#pxd/cdef void gl_init_object_location(GLuint location):
def gl_init_object_location(location):
    cube.object_location = location
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import unicodedata

import polib

try:
    from checkpo_override import version
except ImportError:
    version = 0
if version > 0:
    from checkpo_override import overrides
    
override_filename = os.path.dirname(__file__)+'/checkpo_override.py'

override_defaults = dict(rpat=None, extrachars='', transchars=())

def overrides_to_internal():
    if version == 0:
        if os.path.exists(override_filename):
            raise Exception('{!r} exists, but version is 0'.format(override_filename))
        return []
    if version == 3:
        def to_key(item):
            return item['lang'], item['msgid'], item['msgstr']
        def to_val(item):
            def getitem(key):
                try:
                    return item[key]
                except KeyError:
                    return override_defaults[key]
            return item['opat'], getitem('rpat'), getitem('extrachars'), getitem('transchars'), 'unused'
        return {to_key(item):to_val(item) for item in overrides}
    raise Exception('unknown version {} in {!r}'.format(version, override_filename))
overrides = overrides_to_internal()

def write_overrides(overrides):
    from io import StringIO
    from pprint import pprint, pformat
    def overrides_sort(val):
        lang, linenum, msgid, msgstr, *unused = val
        if isinstance(linenum, int):
            assert type(lang) is str, val
            assert type(linenum) is int, val
            return lang, '', linenum, msgstr
        else:
            assert type(lang) is str, val
            assert type(linenum) is str, val
            return lang, linenum, 0, msgstr
    overrides_gen = ((l,n,mi,ms,op,rp,ec,rc) for (l,mi,ms),(op,rp,ec,rc,n) in overrides.items())
    overrides_ = sorted(overrides_gen, key=overrides_sort)
    out = StringIO()
    
    def piprint(prefix, msg, out):
        msg = pformat(msg, indent=1, width=100)
        msg = ('\n'+' '*(len(prefix)+7)).join(msg.splitlines())
        print('      ' + prefix + '=' + msg + ',', file=out)
    print('version = 3', file=out)
    print('overrides = [', file=out)
    for l,n,mi,ms,op,rp,ec,rc in overrides_:
        print('dict( lang={!r}, linenum={},'.format(l, n), file=out)
        vals = zip(('msgid ','msgstr','opat','rpat','extrachars','transchars'),
                   (mi,      ms,      op,    rp,    ec,          rc))
        vals = ((n,v) for n,v in vals if n not in override_defaults or override_defaults[n] != v)
        for name, val in vals:
            piprint(name, val, out)
        print('    ),', file=out)
    print(']', file=out)
    print(file=out)
    
    out = out.getvalue()
    with open(override_filename, 'wt', encoding='utf-8') as overridefile:
        overridefile.write(out)
        
        
extra_chars = "-,;'/"
trans_chars = ('…...',':.',':;','"«','"»')
extra_chars_lang = {'ast': "¡", 'bn': '/', 'es': '¡´', 'fr': '\xa0’', 'gl': '\u0303', 'he': '\u05be\u05f3\u200f',
                    'ja': '\u30fc\u3001', 'km': '\xa0\u200b', 'lo': '\u0ec6', 'ru': '—', 'uk': '’—', 'uz': '‘ʻ’',
                    'zh_CN': '\uff0c\uff1b', 'zh_TW': '\uff0c\uff1b',}
trans_chars_lang = {'bn': ('.\u0964',),
                    'cs': ('×x',), 'de': ('"„', '"“'), 'fr': ('×x',), 'gl': ('×x',), 'he': ('"„', '"“'),
                    'it': ('–-', '×x'), 'ja': ('!\uff01',), 'km': ('.\u17d4',), 'ms': ('–-', '×x'),
                    'pl': ('×x',), 'ru': ('–—',), 'uk': ('–—',),
                    'zh_CN': (':\uff1a', '!\uff01', '.\u3002', ':\uff0c', '"“', '"”'),
                    'zh_TW': (':\uff1a', '.\u3002', ),
                    }
extra_ucats_lang = {'bn': ('Lo', 'Mc', 'Mn'), 'he': ('Lo',), 'ja': ('Lo',), 'km': ('Lo', 'Mc', 'Mn'),
                    'lo' : ('Lo', 'Mn'), 'te': ('Lo', 'Mc', 'Mn'),
                    'th': ('Lo', 'Mn'), 'zh_CN': ('Lo',), 'zh_TW': ('Lo',),
                    }
#TODO: special checks for msgids like data/applications/pybik.desktop.in.h:4


class MsgidError (Exception):
    def __str__(self):
        return 'pot: {} in {!r}'.format(*self.args)
class MsgError (Exception):
    def __str__(self):
        return '{}: {} in {!r}'.format(*self.args)
        
def ischar(lang, c):
    return unicodedata.category(c) in ['Ll', 'Lu'] or c in extra_chars
    
def ischarsp(lang, c, extrachars):
    if c == ' ':
        return True
    if c in extrachars:
        return True
    if ischar(lang, c):
        return True
    return unicodedata.category(c) in extra_ucats_lang.get(lang, ()) or c in extra_chars_lang.get(lang, '')
    
def extract_pattern(lang, msgid):
    def ep_gen():
        maybebrace = False
        nbraces = 0
        bracestr = ''
        atstart = True
        spaces = ''
        link = ''
        digits = ''
        
        c = None
        imsgid = iter(msgid)
        for c in imsgid:
            if maybebrace:
                maybebrace = False
                if c == '{':
                    yield '{{'
                elif c == '}':
                    yield '{}'
                else:
                    nbraces = 1
                    bracestr = '{' + c
                continue
            elif nbraces:
                if c == '\n': raise MsgidError('newline inside {}', msgid)
                if c == '{':  nbraces += 1
                if c == '}':  nbraces -= 1
                bracestr += c
                if not nbraces:
                    yield bracestr
                    bracestr = ''
                continue
            elif link:
                if c == '\n': raise MsgidError('newline inside <>', msgid)
                link += c
                if c == '>':
                    yield link
                    link = ''
                continue
            elif digits:
                if c in '0123456789':
                    digits += c
                    continue
                yield digits
                digits = ''
            elif spaces:
                if c == ' ':
                    spaces += c
                    continue
                elif atstart and c == '*':
                    spaces += c
                    continue
                elif c == '\n':
                    atstart = True
                    spaces += c
                    continue
                else:
                    if atstart:
                        yield spaces
                        atstart = False
                    spaces = ''
            if c == '\n':
                atstart = True
                spaces = c
                continue
            if c == ' ':
                spaces = c
                continue
            elif c == '{':
                maybebrace = True
            elif c == '<':
                link = c
            elif c in '0123456789':
                digits = c
            elif not ischar(lang, c):
                yield c
            atstart = False
        if maybebrace:  raise MsgidError('unclosed { at end', msgid)
        elif nbraces:   raise MsgidError('unclosed { at end', msgid)
        elif link:      raise MsgidError('unclosed < at end', msgid)
        elif digits:    yield digits
        elif spaces:    yield spaces
    return tuple(ep_gen())
    
def startswith_pattern(lang, string, pattern, replacechars):
    if string.startswith(pattern):
        return pattern
    for co, *cr in replacechars:
        cr = ''.join(cr)
        if co == pattern and string.startswith(cr):
            return cr
    for co, *cr in trans_chars_lang.get(lang, ()):
        cr = ''.join(cr)
        if co == pattern and string.startswith(cr):
            return cr
    for co, *cr in trans_chars:
        cr = ''.join(cr)
        if co == pattern and string.startswith(cr):
            return cr
    return None
    
def match_pattern(lang, entry, patterns, msgstr):
    msgrest = msgstr
    assert type(msgrest) is str, type(msgrest)
    opatterns, rpatterns, extrachars, replacechars, linenum = overrides.get(
                            (lang, entry.msgid, msgstr), (patterns, None, '', (), 'standard'))
    if opatterns != patterns:
        print('{}:error: pattern {} changed to {}'.format(entry.linenum, opatterns, patterns))
        opatterns = patterns
    if linenum == 'unused':
        if (rpatterns is None or opatterns == rpatterns) and not extrachars and replacechars == ():
            del overrides[(lang, entry.msgid, msgstr)]
        else:
            overrides[(lang, entry.msgid, msgstr)] = opatterns, rpatterns, extrachars, replacechars, entry.linenum
    else:
        assert linenum == 'standard' or linenum == entry.linenum, (linenum, entry.linenum)
    if rpatterns is not None:
        patterns = rpatterns
    for i, pattern in enumerate(patterns):
        while msgrest:
            rpattern = startswith_pattern(lang, msgrest, pattern, replacechars)
            if rpattern is not None:
                msgrest = msgrest[len(rpattern):]
                break
            else:
                c = msgrest[0]
                if not ischarsp(lang, c, extrachars):
                    overrides[(lang, entry.msgid, msgstr)] = opatterns, rpatterns, extrachars, replacechars, entry.linenum
                    print('{}:error: found {!r} ({} {}) while searching for pattern[{}] of {!r}'.format(
                            entry.linenum, c, hex(ord(c)), unicodedata.category(c), i, patterns))
                    break
                msgrest = msgrest[1:]
        else:
            rpattern = startswith_pattern(lang, msgrest, pattern, replacechars)
            if rpattern is None:
                overrides[(lang, entry.msgid, msgstr)] = opatterns, rpatterns, extrachars, replacechars, entry.linenum
                print('{}:error: pattern[{}] {!r} not found'.format(entry.linenum, i, pattern))
    while msgrest:
        c = msgrest[0]
        if not ischarsp(lang, c, extrachars):
            overrides[(lang, entry.msgid, msgstr)] = opatterns, rpatterns, extrachars, replacechars, entry.linenum
            print('{}:error: found {!r} ({} {}) while searching for end'.format(
                    entry.linenum, c, hex(ord(c)), unicodedata.category(c)))
            break
        msgrest = msgrest[1:]
        
def check_postring(lang, entry):
    if not entry.translated():
        return
    pattern = extract_pattern(lang, entry.msgid)
    if entry.msgid_plural:
        pattern_plural = extract_pattern(lang, entry.msgid_plural)
        if pattern != pattern_plural:
            raise MsgError(lang, 'msgid != msgid_plural', entry.msgid)
        if entry.msgstr:
            raise MsgError(lang, 'msgstr not empty for plural', entry.msgstr)
        for msgstr in entry.msgstr_plural.values():
            match_pattern(lang, entry, pattern, msgstr)
    else:
        if entry.msgstr_plural:
            raise MsgError(lang, 'msgstr_plural not empty for singular', entry.msgstr_plural)
        match_pattern(lang, entry, pattern, entry.msgstr)
        
        
def usage():
    prog = os.path.basename(__file__)
    print('usage:')
    print(' ', prog, '-h')
    print(' ', prog)
    print('      print errors about:', override_filename)
    print(' ', prog, '-u')
    print('      write/update', override_filename)
    
def main():
    args = sys.argv[1:]
    update = False
    for arg in args:
        if arg == '-h':
            usage()
            return
        if arg == '-u':
            update = True
        else:
            print('unknown argument:', arg)
            usage()
            return
            
    try:
        for filename in sorted(os.listdir('po')):
            if filename.endswith('.po'):
                print('processing', filename)
                for entry in polib.pofile(os.path.join('po', filename)):
                    check_postring(os.path.splitext(filename)[0], entry)
    except PipeError as e:
        print(e)
        return
    if update:
        write_overrides(overrides)
        
        
if __name__ == '__main__':
    main()
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


def lines():
    template_1 = '''
Fields:
  solved = [True]
  type = ['Cube']
  sizes = [(3, 3, 3)]
  selection_mode = ['quadrant', 'simple']
  mirror_faces = [False, True]
Conditions:
Limits:
Initial:
  State: solved=True, type='Cube', sizes=(3, 3, 3), selection_mode='simple', mirror_faces=False
Transition: set_settings_draw 'selection_nick', 'quadrant'
  State:
         selection_mode='quadrant'
Transition: set_settings_draw 'selection_nick', 'simple'
  State:
         selection_mode='simple'
Transition: set_settings_draw 'mirror_faces', False
  State:
         mirror_faces=False
Transition: set_settings_draw 'mirror_faces', True
  State:
         mirror_faces=True
Transition: direct_selectmodel 'Cube', (3,)
'''.strip()
    template_2 = '''
            direct_selectmodel {!r}, {!r}
'''.strip('\n')
    template_3 = '''
  State: solved=True, type='Cube', sizes=(3, 3, 3)
         solved=True, type='Cube', sizes=(3, 3, 3)
'''.strip('\n')
    template_4 = '''
                      type={!r}, sizes={!r}
'''.strip('\n')
    
    from pybiklib.model import Model
    Model.load_index()
    
    tslist = []
    for mtype, sizesdict in Model.cache_index['sizes'].items():
        for normsizes, (internsizes, unused_file) in sizesdict.items():
            tslist.append((mtype, normsizes, internsizes))
    import random
    #tslist.sort()
    random.shuffle(tslist)
    tslist.append(('Cube', (3,), (3, 3, 3)))
            
    yield from template_1.splitlines()
    for mtype, normsizes, internsizes in tslist:
        yield from template_2.format(mtype, normsizes).splitlines()
    yield from template_3.splitlines()
    for mtype, normsizes, internsizes in tslist:
        yield from template_4.format(mtype, internsizes).splitlines()
        
def main():
    import sys
    sys.path.insert(0, '.')
    for line in lines():
        print(line)
        
    
if __name__ == '__main__':
    main()
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
sys.path.insert(0, '.')
from glob import glob
from subprocess import call, Popen, PIPE
from difflib import unified_diff
import pickle
import pickletools
from pprint import pformat
from io import BytesIO, StringIO

from pybiklib.config import MODELS_DIR
from pybiklib.model import Model


pformat_kwargs = {'compact': False}
try:
    pformat_kwargs['width'] = os.environ['COLUMNS']
except KeyError:
    pass
    
    
def pager(lines):
    process = Popen('less', stdin=PIPE, universal_newlines=True)
    try:
        process.stdin.writelines(lines)
    except BrokenPipeError:
        pass
    else:
        process.stdin.close()
    finally:
        process.wait()
        
def bindiff(bytes1, bytes2):
    for i, (a, b) in enumerate(zip(bytes1, bytes2)):
        if a != b:
            print('      first difference in byte', i+1)
            break
            
def pickle_disassembly(bytes):
    lines = []
    strio = StringIO()
    pickletools.dis(bytes, out=strio, annotate=1)
    lines += strio.getvalue().splitlines(keepends=True)
    #for opcode, arg, pos in pickletools.genops(bytes):
    #    lines.append('{2:4}: {0.name} arg:{1!r}\n'.format(opcode, arg, pos))
    #    lines.append('      {0.doc}\n'.format(opcode))
    return lines
    
def show_diff(filename1, filename2, bytes1, bytes2):
    len1 = len(bytes1)
    len2 = len(bytes2)
    if len1 != len2:
        print('    file contents have different length:', len1, len2)
    if bytes1 != bytes2:
        print('    file contents are different')
        bindiff(bytes1, bytes2)
        
    buffer1 = BytesIO(bytes1)
    data1 = pickle.load(buffer1)
    text1 = pformat(data1, **pformat_kwargs)
    dlen1 = buffer1.tell()
    vlen1 = len1 - dlen1
    vectors1 = Model.read_vectors(buffer1) if vlen1 else []
    
    buffer2 = BytesIO(bytes2)
    data2 = pickle.load(buffer2)
    text2 = pformat(data2, **pformat_kwargs)
    dlen2 = buffer2.tell()
    vlen2 = len2 - dlen2
    vectors2 = Model.read_vectors(buffer2) if vlen2 else []
    
    lines = []
    
    if dlen1 != dlen2:
        print('    data parts have different length:', dlen1, dlen2)
    if bytes1[:dlen1] != bytes2[:dlen2]:
        print('    data parts are different')
        bindiff(bytes1[:dlen1], bytes2[:dlen2])
        dis1 = pickle_disassembly(bytes1[:dlen1])
        dis2 = pickle_disassembly(bytes2[:dlen2])
        lines += unified_diff(dis1, dis2, filename1, filename2, 'pickle disassembly')
    if text1 != text2:
        print('    data parts have different content')
        lines += unified_diff(text1.splitlines(), text2.splitlines(), filename1, filename2, 'data')
    if vlen1 != vlen2:
        print('    vector parts have different length:', vlen1, vlen2)
    if bytes1[dlen1:] != bytes2[dlen2:]:
        print('    vector parts are different')
        bindiff(bytes1[dlen1:], bytes2[dlen2:])
        vectors1 = [str(v)+'\n' for v in vectors1]
        vectors2 = [str(v)+'\n' for v in vectors2]
        lines += unified_diff(vectors1, vectors2, filename1, filename2, 'vectors')
        
    if lines:
        pager(lines)
        
def build_models(args):
    call('./setup.py build_models -i --reproducible'.split() + args)
    
def compare_models(dirname1, dirname2):
    assert os.path.exists(dirname1)
    assert os.path.exists(dirname2)
    filenames1 = os.listdir(dirname1)
    filenames2 = os.listdir(dirname2)
    print('comparing', dirname1, 'and', dirname2)
    for filename in sorted(set(filenames1 + filenames2)):
        if filename not in filenames1:
            print(' ', filename, 'not in', dirname1)
        elif filename not in filenames2:
            print(' ', filename, 'not in', dirname2)
        else:
            fn1 = os.path.join(dirname1, filename)
            fn2 = os.path.join(dirname2, filename)
            with open(fn1, 'rb') as f1, open(fn2, 'rb') as f2:
                s1 = f1.read()
                s2 = f2.read()
            if s1 != s2:
                print(' ', filename)
                show_diff(fn1, fn2, s1, s2)
                
def main():
    args = sys.argv[1:]
    if not args:
        print('usage:', os.path.basename(sys.argv[0]), 'n [buildargs]')
        print('build models n times and check whether they are the same')
        return
    count = int(args[0])
    args = args[1:]
    prevdirname = None
    i = -1
    for i, dirname in enumerate(sorted(glob(MODELS_DIR + '-*'))):
        curdirname = MODELS_DIR + '-{:04}'.format(i + 1)
        if curdirname != dirname and not os.path.exists(curdirname):
            os.rename(dirname, curdirname)
        else:
            curdirname = dirname
        if prevdirname is not None:
            compare_models(prevdirname, curdirname)
        prevdirname = curdirname
    if os.path.exists(MODELS_DIR) and prevdirname is not None:
        compare_models(prevdirname, MODELS_DIR)
    i += 1
    for unused_i in range(i, i + count):
        if os.path.exists(MODELS_DIR):
            i += 1
            prevdirname = MODELS_DIR + '-{:04}'.format(i)
            os.rename(MODELS_DIR, prevdirname)
        build_models(args)
        if not os.path.exists(MODELS_DIR):
            print('build failed')
            break
        if prevdirname is not None:
            compare_models(prevdirname, MODELS_DIR)
        
if __name__ == '__main__':
    main()
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2011-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os, re
sys.path.insert(0, os.getcwd())

from pybiklib import model
model.Model.load_index()
from pybiklib import pluginlib
pluginlib._ = lambda v: '_({!r})'.format(v)
pluginlib.ngettext = lambda s,p,n: 'ngettext({!r}, {!r}, {})'.format(s,p,n).replace('{','{{').replace('}','}}')

try:
    from send_tracebacks import send_dbus_message
except ImportError:
    send_dbus_message = lambda *args: print(*args)
    

def convert(source_filename, target_filename):
    plugin = pluginlib.PluginFileLoader(source_filename)
    source_filename = os.path.abspath(source_filename)
    with open(target_filename, 'wt', encoding='utf-8') as target:
        translation_comment = None
        dstlineno = 0
        srclineno = 0
        for paragraph in plugin.body:
            path, pathitems, srclineno, translation_comment = paragraph['Path']
            if translation_comment is not None:
                if ':' in translation_comment:
                    translation_key, translation_comment = translation_comment.strip('#').strip().split(':', 1)
                    if translation_key:
                        translation_comment = '# ' + translation_comment.strip()
                    else:
                        translation_comment = None
                        send_dbus_message('Warning', 'Empty translation key', source_filename, srclineno)
                else:
                    translation_key = translation_comment.strip('#').strip().split()
                    if translation_key:
                        translation_key = translation_key[0]
                    else:
                        translation_comment = None
                        send_dbus_message('Warning', 'Empty translation key', source_filename, srclineno)
                if ' ' in translation_key:
                    send_dbus_message('Warning', 'Translation key contains whitespace', source_filename, srclineno)
            if pathitems is not None:
                values = []
                for unused, v in pathitems:
                    if v:
                        if translation_comment and translation_key in re.split(r'\W+', v):
                            values.append(translation_comment)
                        values.append(v)
                while dstlineno < srclineno:
                    target.write('\n')
                    dstlineno += 1
                for v in values:
                    target.write(v)
                    target.write('\n')
                    dstlineno += 1
                translation_comment = None
            else:
                translation_comment = None
                continue
    
def main():
    if len(sys.argv) != 2:
        print('usage: {} scriptfile'.format(os.path.basename(sys.argv[0])))
        return 1
    filename = sys.argv[1]
    convert(filename, filename+'.py')
    
if __name__ == '__main__':
    sys.exit(main())


